"""
Structure for parsed as dict request or response syntax values.
"""
from __future__ import annotations

from typing import Dict, Any, List, Optional

from mypy_boto3_builder.type_maps.syntax_type_map import SYNTAX_TYPE_MAP
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_annotation import TypeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.logger import get_logger


class TypeValue:
    """
    Structure for parsed as dict request or response syntax values.
    """

    def __init__(self, prefix: str, value: Dict[str, Any]) -> None:
        self.logger = get_logger()
        self.prefix = prefix
        self.raw: Dict[str, Any] = value
        self.dict_items: Optional[List[Dict[str, Any]]] = value.get("dict_items")
        if value.get("empty_dict"):
            self.dict_items = []
        self.set_items: Optional[List[Any]] = value.get("set_items")
        self.list_items: Optional[List[Any]] = value.get("list_items")
        if value.get("empty_list"):
            self.list_items = []
        self.func_call: Optional[Dict[str, Any]] = value.get("func_call")
        self.union_items: List[Any] = []
        if value.get("union_first_item"):
            self.union_items.append(value["union_first_item"])
            self.union_items.extend(value["union_rest_items"])
        self.literal_items: List[Any] = []
        if value.get("literal_first_item"):
            self.literal_items.append(value["literal_first_item"])
            self.literal_items.extend(value["literal_rest_items"])

        self.value: Optional[str] = value.get("value")

    def is_dict(self) -> bool:
        return self.dict_items is not None

    def is_list(self) -> bool:
        return self.list_items is not None

    def is_literal(self) -> bool:
        return bool(self.literal_items)

    def is_set(self) -> bool:
        return bool(self.set_items)

    def is_union(self) -> bool:
        return bool(self.union_items)

    def is_func_call(self) -> bool:
        return bool(self.func_call)

    def is_plain(self) -> bool:
        return self.value is not None

    def _get_type_dict(self) -> FakeAnnotation:
        if not self.dict_items:
            return TypeSubscript(Type.Dict, [Type.str, Type.Any])

        first_key = self.dict_items[0]["key"]
        if first_key in SYNTAX_TYPE_MAP:
            result = TypeSubscript(Type.Dict)
            result.add_child(SYNTAX_TYPE_MAP[first_key])
            result.add_child(
                TypeValue(self.prefix, self.dict_items[0]["value"]).get_type()
            )
            return result

        typed_dict = TypeTypedDict(f"{self.prefix}TypeDef")
        for item in self.dict_items:
            key_name = self._parse_constant(item["key"])
            prefix = f"{self.prefix}{key_name}"
            typed_dict.add_attribute(
                key_name, TypeValue(prefix, item["value"]).get_type(), required=False,
            )
        return typed_dict

    def _get_type_list(self) -> TypeSubscript:
        if not self.list_items:
            return TypeSubscript(Type.List, [Type.Any])

        result = TypeSubscript(Type.List)
        for item_index, item in enumerate(self.list_items):
            prefix = f"{self.prefix}{item_index if item_index else ''}"
            result.add_child(TypeValue(prefix, item).get_type())
        return result

    def _get_type_union(self) -> FakeAnnotation:
        if not self.union_items:
            return Type.Any

        result = TypeSubscript(Type.Union)
        for item_index, item in enumerate(self.union_items):
            prefix = f"{self.prefix}{item_index if item_index else ''}"
            result.add_child(TypeValue(prefix, item).get_type())

        if all(i is result.children[0] for i in result.children):
            return result.children[0]

        return result

    def _get_type_set(self) -> TypeAnnotation:
        if not self.set_items:
            return Type.Any

        plain_values = [i["value"] for i in self.set_items]
        if plain_values == ["'... recursive ...'"]:
            return Type.Any

        self.logger.warning(f"Unknown set: {self.raw}, fallback to Any")
        return Type.Any

    def _get_type_func_call(self) -> FakeAnnotation:
        if not self.func_call:
            raise ValueError(f"Value is not a func call: {self.raw}")

        if self.func_call["name"] == "datetime":
            return ExternalImport(ImportString("datetime"), "datetime")

        if self.func_call["name"] == "StreamingBody":
            return ExternalImport(ImportString("botocore", "response"), "StreamingBody")

        if self.func_call["name"] == "EventStream":
            return ExternalImport(
                ImportString("botocore", "eventstream"), "EventStream"
            )

        self.logger.warning(f"Unknown function: {self.raw}, fallback to Any")
        return Type.Any

    def _get_type_plain(self) -> FakeAnnotation:
        if not self.value:
            raise ValueError(f"Value is not plain: {self.raw}")

        if self.value in SYNTAX_TYPE_MAP:
            return SYNTAX_TYPE_MAP[self.value]

        if self.value.startswith("'"):
            return Type.str

        self.logger.warning(f"Unknown plain value: {self.raw}, fallback to Any")
        return Type.Any

    def is_literal_item(self) -> bool:
        if self.value is None:
            return False
        return self.value.startswith("'")

    def _get_type_literal(self) -> FakeAnnotation:
        if not self.literal_items:
            raise ValueError(f"Value is not literal: {self.raw}")

        items = [TypeValue(self.prefix, i) for i in self.literal_items]
        if all(i.is_literal_item() for i in items):
            item_constants = [self._parse_constant(i.value or "") for i in items]
            return TypeLiteral(*item_constants)

        item_types = [i.get_type() for i in items]

        if all([i is item_types[0] for i in item_types]):
            return item_types[0]

        return TypeSubscript(Type.Union, item_types)

    @staticmethod
    def _parse_constant(value: str) -> Any:
        if value.startswith("'"):
            return value.replace("'", "")

        if value.isdigit():
            return int(value)

        raise ValueError(f"Invalid constant: {value}")

    def get_type(self) -> FakeAnnotation:
        if self.is_list():
            return self._get_type_list()
        if self.is_dict():
            return self._get_type_dict()
        if self.is_set():
            return self._get_type_set()
        if self.is_func_call():
            return self._get_type_func_call()
        if self.is_union():
            return self._get_type_union()
        if self.is_literal():
            return self._get_type_literal()
        if self.is_plain():
            return self._get_type_plain()

        raise ValueError(f"Unknown value: {self.raw}")



"""
Module-level function.
"""
from dataclasses import dataclass, field
from typing import Set, List

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.structures.argument import Argument


@dataclass
class Function:
    """
    Module-level function.
    """

    name: str
    arguments: List[Argument]
    return_type: FakeAnnotation
    docstring: str = ""
    decorators: List[FakeAnnotation] = field(default_factory=lambda: [])
    body_lines: List[str] = field(default_factory=lambda: [])

    @property
    def body(self) -> str:
        return "\n".join(self.body_lines)

    def get_types(self) -> Set[FakeAnnotation]:
        types = self.return_type.get_types()
        for argument in self.arguments:
            types.update(argument.get_types())
        for decorator in self.decorators:
            types.update(decorator.get_types())

        return types

    def get_required_import_records(self) -> Set[ImportRecord]:
        result: Set[ImportRecord] = set()
        for type_annotation in self.get_types():
            import_record = type_annotation.get_import_record()
            if not import_record:
                continue
            if import_record.is_builtins():
                continue
            result.add(import_record)

        return result



"""
Parser for Boto3 ServiceResource sub-resource, produces `structures.Resource`
"""
import inspect
from typing import Dict, Type
from types import FunctionType


from boto3.docs.utils import is_resource_action
from boto3.resources.base import ServiceResource as Boto3ServiceResource

from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.resource import Resource
from mypy_boto3_builder.parsers.helpers import parse_attributes, parse_method
from mypy_boto3_builder.parsers.identifiers import parse_identifiers
from mypy_boto3_builder.parsers.parse_collections import parse_collections
from mypy_boto3_builder.parsers.shape_parser import ShapeParser


def parse_resource(
    name: str,
    resource: Boto3ServiceResource,
    service_name: ServiceName,
    shape_parser: ShapeParser,
) -> Resource:
    """
    Parse boto3 sub Resource data.

    Arguments:
        resource -- Original boto3 resource.

    Returns:
        Resource structure.
    """
    result = Resource(
        name=name,
        docstring=(
            f"[{name} documentation]"
            f"({service_name.doc_link}.ServiceResource.{name})"
        ),
    )
    shape_methods_map = shape_parser.get_resource_method_map(name)
    public_methods = get_resource_public_methods(resource.__class__)
    for method_name, public_method in public_methods.items():
        if method_name in shape_methods_map:
            method = shape_methods_map[method_name]
        else:
            method = parse_method(name, method_name, public_method)
        result.methods.append(method)

    result.attributes.extend(parse_attributes(resource))
    result.attributes.extend(parse_identifiers(resource))

    collections = parse_collections(name, resource, service_name)
    for collection in collections:
        result.collections.append(collection)
        result.attributes.append(
            Attribute(
                collection.attribute_name, InternalImport(collection.name, service_name)
            )
        )

    return result


def get_resource_public_methods(
    resource_class: Type[Boto3ServiceResource],
) -> Dict[str, FunctionType]:
    """
    Extract public methods from boto3 sub resource.

    Arguments:
        resource_class -- boto3 resource meta.

    Returns:
        A dictionary of method name and method.
    """
    class_members = inspect.getmembers(resource_class)
    methods: Dict[str, FunctionType] = {}
    for name, member in class_members:
        if name.startswith("_"):
            continue

        if name[0].isupper():
            continue

        if not is_resource_action(member):
            continue

        methods[name] = member

    return methods



"""
String to type annotation map that find type annotation by method and argument name.
"""
from typing import Dict

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_maps.typed_dicts import ec2_tag_type, waiter_config_type


METHOD_TYPE_MAP: Dict[str, FakeAnnotation] = {
    "page_size: count": Type.int,
    "limit: count": Type.int,
    "create_tags: Resources": TypeSubscript(Type.List, [Type.Any]),
    "create_tags: Tags": TypeSubscript(Type.List, [ec2_tag_type]),
    "create_tags: DryRun": Type.bool,
    "wait: WaiterConfig": waiter_config_type,
}



"""
Base class for all structures that can be rendered to a class.
"""
from dataclasses import dataclass, field
from typing import Set, List

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.import_helpers.import_record import ImportRecord


@dataclass
class ClassRecord:
    """
    Base class for all structures that can be rendered to a class.
    """

    name: str
    methods: List[Method] = field(default_factory=lambda: [])
    attributes: List[Attribute] = field(default_factory=lambda: [])
    bases: List[FakeAnnotation] = field(default_factory=lambda: [])
    docstring: str = ""

    def get_types(self) -> Set[FakeAnnotation]:
        types: Set[FakeAnnotation] = set()
        for method in self.methods:
            types.update(method.get_types())
        for attribute in self.attributes:
            types.update(attribute.get_types())
        for base in self.bases:
            types.update(base.get_types())
        return types

    def get_required_import_records(self) -> Set[ImportRecord]:
        result: Set[ImportRecord] = set()
        for type_annotation in self.get_types():
            import_record = type_annotation.get_import_record()
            if not import_record:
                continue
            if import_record.is_builtins():
                continue
            result.add(import_record)

        return result



"""
Parent class for all package structures.
"""

from dataclasses import dataclass


@dataclass
class Package:
    """
    Parent class for all package structures.
    """

    name: str
    pypi_name: str



"""
Pyparsing grammar for request and response syntax.
"""
from pyparsing import (
    Word,
    alphanums,
    Group,
    alphas,
    SkipTo,
    Optional,
    delimitedList,
    Combine,
    Forward,
    Literal,
    ParserElement,
)


class SyntaxGrammar:
    """
    ellipsis = "..."
    variable_name ::= alphanums + "_-."
    name_value ::= alphanums + "_-."
    string_value ::= alphas{0,2} "'"  [^']+  "'"
    plain_value ::= string_value | name_value
    literal_item ::= list_value | dict_value | set_value | plain_value
    literal_value ::= literal_item ("|" literal_item)+
    any_value ::= literal_value | list_value | dict_value | set_value | union_value | func_call | plain_value
    empty_list_value ::= "[" [ellipsis] [","] "]"
    non_empty_list_value ::= "[" any_value ("," any_value)* [","] "]"
    list_value ::= empty_list_value | non_empty_list_value
    set_value ::= "{" any_value ("," any_value)* [","] "}"
    func_call ::= name_value "(" any_value ("," any_value)* [","] ")"
    empty_dict_value ::= "{" [ellipsis] [","] "}"
    non_empty_dict_value ::= "{" string_value ":" any_value ("," string_value ":" any_value)* [","] "}"
    dict_value ::= empty_dict_value | non_empty_dict_value
    union_item ::= literal_value | list_value | dict_value | set_value | plain_value
    union_value ::= union_item ("or" union_item)+
    argument ::= alphanums "=" any_value
    definition ::= [^']+ "(" argument ("," argument)* [","] ")"
    request_syntax ::= "**Request Syntax**" "::" definition
    response_syntax ::= "**Response Syntax**" "::" (list_value | dict_value)
    """

    ellipsis = Literal("...")
    variable_name = Word(alphanums + "_-.")
    name_value = Word(alphanums + "_-.")
    string_value = Combine(
        Optional(Word(alphas, max=2)) + Literal("'") + SkipTo("'") + Literal("'")
    )
    plain_value = (string_value | name_value).setResultsName("value")
    literal_item = Forward()
    literal_value = (
        Group(literal_item).setResultsName("literal_first_item")
        + Literal("|")
        + delimitedList(Group(literal_item), delim="|").setResultsName(
            "literal_rest_items"
        )
    )
    any_value = Forward()
    empty_list_value = Group(
        Literal("[") + Optional(ellipsis) + Optional(",") + Literal("]")
    ).setResultsName("empty_list")
    non_empty_list_value = (
        Literal("[")
        + Group(delimitedList(Group(any_value))).setResultsName("list_items")
        + Optional(",")
        + Literal("]")
    )
    list_value = empty_list_value | non_empty_list_value
    set_value = (
        "{"
        + Group(delimitedList(Group(any_value))).setResultsName("set_items")
        + Optional(",")
        + "}"
    )
    func_call = Group(
        name_value.setResultsName("name")
        + Literal("(")
        + Optional(delimitedList(Group(any_value))).setResultsName("args")
        + Optional(",")
        + Literal(")")
    ).setResultsName("func_call")
    empty_dict_value = Group(
        Literal("{") + Optional(ellipsis) + Optional(",") + Literal("}")
    ).setResultsName("empty_dict")
    non_empty_dict_value = (
        Literal("{")
        + Optional(
            delimitedList(
                Group(
                    string_value.setResultsName("key")
                    + Literal(":").suppress()
                    + Combine(any_value).setResultsName("value")
                )
            ).setResultsName("dict_items")
        )
        + Optional(",")
        + Literal("}")
    )
    dict_value = empty_dict_value | non_empty_dict_value
    literal_item <<= list_value | dict_value | set_value | plain_value
    union_item = literal_value | list_value | dict_value | set_value | plain_value
    union_value = (
        Group(union_item).setResultsName("union_first_item")
        + Literal("or").suppress()
        + delimitedList(Group(union_item), delim="or",).setResultsName(
            "union_rest_items"
        )
    )
    any_value <<= (
        literal_value
        | list_value
        | dict_value
        | set_value
        | union_value
        | func_call
        | plain_value
    )
    argument = (
        Word(alphanums).setResultsName("name")
        + Literal("=")
        + Group(any_value).setResultsName("value")
    )
    definition = (
        SkipTo("(")
        + Literal("(")
        + Optional(delimitedList(Group(argument))).setResultsName("arguments")
        + Optional(",")
        + Literal(")")
    )
    request_syntax = "**Request Syntax**" + Literal("::") + definition
    response_syntax = (
        "**Response Syntax**"
        + Literal("::")
        + Group(list_value | dict_value).setResultsName("value")
    )

    @classmethod
    def reset(cls) -> None:
        cls.disable_packrat()

    @staticmethod
    def enable_packrat() -> None:
        ParserElement.enablePackrat(cache_size_limit=128)

    @staticmethod
    def disable_packrat() -> None:
        ParserElement.enablePackrat(cache_size_limit=None)



"""
Boto3 ServiceResource.
"""
from dataclasses import dataclass, field
from typing import List, Set, Optional

from boto3.resources.base import ServiceResource as Boto3ServiceResource

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.resource import Resource


@dataclass
class ServiceResource(ClassRecord):
    """
    Boto3 ServiceResource.
    """

    name: str = "ServiceResource"
    service_name: ServiceName = ServiceNameCatalog.ec2
    boto3_service_resource: Optional[Boto3ServiceResource] = None
    collections: List[Collection] = field(default_factory=lambda: [])
    sub_resources: List[Resource] = field(default_factory=lambda: [])
    bases: List[FakeAnnotation] = field(
        default_factory=lambda: [
            ExternalImport(
                source=ImportString("boto3", "resources", "base"),
                name="ServiceResource",
                alias="Boto3ServiceResource",
            )
        ]
    )

    def __hash__(self) -> int:
        return hash(self.service_name)

    def get_types(self) -> Set[FakeAnnotation]:
        types = super().get_types()
        for collection in self.collections:
            types.update(collection.get_types())
        for sub_resource in self.sub_resources:
            types.update(sub_resource.get_types())

        return types

    def get_import_records(self) -> Set[ImportRecord]:
        import_records: Set[ImportRecord] = set()
        source = ImportString(
            self.service_name.module_name, ServiceModuleName.service_resource.name
        )

        import_records.add(ImportRecord(source, "ServiceResource"))
        for resource in self.sub_resources:
            import_records.add(ImportRecord(source, resource.name))

        return import_records

    def get_all_names(self) -> List[str]:
        result = [self.name]
        for resource in self.sub_resources:
            result.append(resource.name)
        for collection in self.get_collections():
            result.append(collection.name)
        return result

    def get_collections(self) -> List[Collection]:
        collection_names = [i.name for i in self.collections]
        result: List[Collection] = []
        result.extend(self.collections)
        for resource in self.sub_resources:
            for collection in resource.collections:
                if collection.name in collection_names:
                    raise ValueError(f"Conflicting collections: {collection.name}")
                collection_names.append(collection.name)
                result.append(collection)

        return result



"""
Service package writer.
"""
from pathlib import Path
from typing import List

from mypy_boto3_builder.structures.service_package import ServicePackage
from mypy_boto3_builder.version import __version__ as version
from mypy_boto3_builder.writers.utils import (
    render_jinja2_template,
    blackify,
)
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName


def write_service_package(package: ServicePackage, output_path: Path) -> List[Path]:
    modified_paths: List[Path] = []
    package_path = output_path / package.name

    output_path.mkdir(exist_ok=True)
    package_path.mkdir(exist_ok=True)

    templates_path = Path("service")
    module_templates_path = templates_path / "service"
    file_paths = [
        (output_path / "setup.py", templates_path / "setup.py.jinja2"),
        (output_path / "README.md", templates_path / "README.md.jinja2"),
        (package_path / "version.py", module_templates_path / "version.py.jinja2"),
        (package_path / "__init__.py", module_templates_path / "__init__.py.jinja2"),
        (package_path / "__main__.py", module_templates_path / "__main__.py.jinja2"),
        (package_path / "py.typed", module_templates_path / "py.typed.jinja2"),
        (
            package_path / ServiceModuleName.client.file_name,
            module_templates_path / ServiceModuleName.client.template_name,
        ),
    ]
    if package.service_resource:
        file_paths.append(
            (
                package_path / ServiceModuleName.service_resource.file_name,
                module_templates_path
                / ServiceModuleName.service_resource.template_name,
            )
        )
    if package.paginators:
        file_paths.append(
            (
                package_path / ServiceModuleName.paginator.file_name,
                module_templates_path / ServiceModuleName.paginator.template_name,
            )
        )
    if package.waiters:
        file_paths.append(
            (
                package_path / ServiceModuleName.waiter.file_name,
                module_templates_path / ServiceModuleName.waiter.template_name,
            )
        )
    if package.typed_dicts:
        file_paths.append(
            (
                package_path / ServiceModuleName.type_defs.file_name,
                module_templates_path / ServiceModuleName.type_defs.template_name,
            )
        )

    for file_path, template_path in file_paths:
        content = render_jinja2_template(
            template_path, package=package, service_name=package.service_name,
        )
        content = blackify(content, file_path)

        if not file_path.exists() or file_path.read_text() != content:
            modified_paths.append(file_path)
            file_path.write_text(content)

    return modified_paths



"""
Pyparsing grammar for argument type doc lines.
"""
from pyparsing import (
    Forward,
    SkipTo,
    Literal,
    LineEnd,
    LineStart,
    Word,
    alphanums,
    indentedBlock,
    Optional,
    White,
    ParserElement,
)


class TypeDocGrammar:
    """
    EOL ::= ["\r"] "\n"
    SOL ::= LINE_START
    line ::= [^EOL]+ EOL
    word ::= alphanums + "_"
    indented_block ::= INDENT (line_indented | any_line)
    line_indented ::= any_line indented_block
    type_definition ::= ":type" [^:]+ ":" [^EOL]+
    rtype_definition ::= ":rtype:" [^EOL]+
    returns_definition ::= (":returns:" | ":return:") [^EOL]+
    param_definition ::= ":param" [^:]+ ":" [^EOL]+ EOL [indented_block]
    response_structure ::= "**Response Structure**" line [indented_block]
    typed_dict_key_line ::= "-" "**" word "**" "*(" word ")" "--*" [^EOL]+ + EOL
    type_line ::= "-" "*(" word ")" "--*" [^EOL]+ + EOL
    any_line ::= typed_dict_key_line | type_line | line
    """

    indent_stack = [1]
    SOL = LineStart().suppress()
    EOL = LineEnd().suppress()
    word = Word(alphanums + "_")
    line = SkipTo(LineEnd()) + EOL
    line_indented = Forward()
    any_line = Forward()
    indented_block = indentedBlock(
        line_indented | any_line, indentStack=indent_stack
    ).setResultsName("indented")
    line_indented <<= any_line + indented_block

    type_definition = (
        SOL
        + Literal(":type")
        + SkipTo(":").setResultsName("name")
        + Literal(":")
        + SkipTo(EOL).setResultsName("type_name")
    )

    rtype_definition = (
        SOL + Literal(":rtype:") + SkipTo(EOL).setResultsName("type_name")
    )

    returns_definition = (
        SOL
        + (Literal(":returns:") | Literal(":return:"))
        + SkipTo(EOL).setResultsName("description")
    )

    param_definition = (
        SOL
        + Literal(":param")
        + SkipTo(":").setResultsName("name")
        + Literal(":")
        + SkipTo(EOL).setResultsName("description")
        + EOL
        + Optional(indented_block)
    )

    response_structure = Literal("**Response Structure**") + line_indented

    typed_dict_key_line = (
        Literal("-")
        + White(ws=" \t")
        + Literal("**")
        + word.setResultsName("name")
        + Literal("**")
        + White(ws=" \t")
        + Literal("*(")
        + word.setResultsName("type_name")
        + Literal(")")
        + White(ws=" \t")
        + Literal("--*")
        + SkipTo(EOL).setResultsName("description")
        + EOL
    )

    type_line = (
        Literal("-")
        + White(ws=" \t")
        + Literal("*(")
        + word.setResultsName("type_name")
        + Literal(")")
        + White(ws=" \t")
        + Literal("--*")
        + SkipTo(EOL).setResultsName("description")
        + EOL
    )

    any_line <<= (typed_dict_key_line | type_line | line).setResultsName("line")

    @classmethod
    def fail_action(
        cls, _input_string: str, _chr_index: int, _source: str, error: BaseException
    ) -> None:
        if "found end of text" not in str(error):
            raise error

    @classmethod
    def reset(cls) -> None:
        cls.disable_packrat()
        cls.indented_block.setFailAction(cls.fail_action)
        cls.indent_stack.clear()
        cls.indent_stack.append(1)

    @staticmethod
    def enable_packrat() -> None:
        ParserElement.enablePackrat(cache_size_limit=128)

    @staticmethod
    def disable_packrat() -> None:
        ParserElement.enablePackrat(cache_size_limit=None)



# pylint: disable=unused-argument,multiple-statements,super-init-not-called,no-self-use
from typing import Any, Dict

from botocore.model import ServiceModel
from botocore.waiter import Waiter, WaiterModel
from botocore.session import Session

class ServiceContext:
    def __init__(
        self,
        service_name: str,
        service_model: ServiceModel,
        service_waiter_model: WaiterModel,
        resource_json_definitions: Dict[str, Any],
    ) -> None: ...

def import_module(name: str) -> Any: ...
def lazy_call(full_name: str, **kwargs: Any) -> Any: ...
def inject_attribute(
    class_attributes: Dict[str, Any], name: str, value: Any
) -> None: ...

class LazyLoadedWaiterModel:
    def __init__(self, bc_session: Session, service_name: str, api_version: str): ...
    def get_waiter(self, waiter_name: str) -> Waiter: ...



# pylint: disable=unused-argument,multiple-statements,super-init-not-called
from typing import Any

from botocore.hooks import BaseEventHooks

def inject_delete_tags(event_emitter: BaseEventHooks, **kwargs: Any) -> None: ...
def delete_tags(self: Any, **kwargs: Any) -> None: ...



# pylint: disable=unused-argument,multiple-statements,unused-import
import collections.abc as collections_abc
from typing import Type
from os import PathLike

SOCKET_ERROR: Type[ConnectionError]

def rename_file(current_filename: PathLike, new_filename: PathLike) -> None: ...



# pylint: disable=unused-argument,multiple-statements,super-init-not-called,no-self-use
from __future__ import annotations

import logging
from typing import List, Dict, Any, Optional

from boto3.resources.model import ResourceModel
from botocore.client import BaseClient

logger: logging.Logger

class ResourceMeta:
    def __init__(
        self,
        service_name: str,
        identifiers: Optional[List[str]] = None,
        client: Optional[BaseClient] = None,
        data: Optional[Dict] = None,
        resource_model: Optional[ResourceModel] = None,
    ) -> None:
        self.service_name: str
        self.identifiers: List[str]
        self.client: BaseClient
        self.data: Dict
        self.resource_model: ResourceModel
    def __repr__(self) -> str: ...
    def __eq__(self, other: Any) -> bool: ...
    def copy(self) -> ResourceMeta: ...

class ServiceResource:
    def __init__(
        self, *args: Any, client: Optional[BaseClient] = None, **kwargs: Any
    ) -> None:
        self.meta: ResourceMeta
    def __repr__(self) -> str: ...
    def __eq__(self, other: Any) -> bool: ...
    def __hash__(self) -> int: ...



"""
Wrapper for constant like `False` or `"test"`.
"""
from __future__ import annotations

from typing import Any

from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class TypeConstant(FakeAnnotation):
    """
    Wrapper for constant like `False` or `"test"`.

    Arguments:
        value -- Constant value.
    """

    def __init__(self, value: Any) -> None:
        self.value = value

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if self.value is Ellipsis:
            return "..."

        return repr(self.value)

    def get_import_record(self) -> ImportRecord:
        return ImportRecord.empty()

    def copy(self) -> TypeConstant:
        """
        Create a copy of type annotation wrapper.
        """
        return TypeConstant(self.value)



"""
Description for boto3 service.
"""
from typing import Tuple
from mypy_boto3_builder.constants import PYPI_NAME, MODULE_NAME


__all__ = (
    "ServiceName",
    "ServiceNameCatalog",
)


class ServiceName:
    """
    Description for boto3 service.
    """

    IMPORT_NAMES = {
        "lambda": "lambda_",
    }
    ESSENTIAL = (
        "ec2",
        "rds",
        "s3",
        "lambda",
        "sqs",
        "cloudformation",
        "dynamodb",
    )

    def __init__(self, name: str, class_name: str) -> None:
        self.name = name
        self.import_name = self.IMPORT_NAMES.get(name, name.replace("-", "_"))
        self.class_name = class_name
        self.boto3_version = "latest"

    @property
    def boto3_name(self) -> str:
        return self.name

    @property
    def module_name(self) -> str:
        """
        Package name for given service.
        """
        return f"{MODULE_NAME}_{self.import_name}".rstrip("_")

    @property
    def pypi_name(self) -> str:
        """
        Name of package on PyPI.
        """
        return f"{PYPI_NAME}-{self.name}"

    @property
    def extras_name(self) -> str:
        """
        Extras name for subpackage installation.
        """
        return self.name

    def is_essential(self) -> bool:
        return self.name in self.ESSENTIAL

    @property
    def doc_link(self) -> str:
        return (
            "https://boto3.amazonaws.com/v1/documentation/api/"
            f"{self.boto3_version}/reference/services/{self.boto3_name}.html#{self.class_name}"
        )


class ServiceNameCatalog:
    """
    Finder for boto3 services by name.
    """

    ITEMS: Tuple[ServiceName, ...] = (
        ServiceName("accessanalyzer", "AccessAnalyzer"),
        ServiceName("acm", "ACM"),
        ServiceName("acm-pca", "ACMPCA"),
        ServiceName("alexaforbusiness", "AlexaForBusiness"),
        ServiceName("amplify", "Amplify"),
        ServiceName("apigateway", "APIGateway"),
        ServiceName("apigatewaymanagementapi", "ApiGatewayManagementApi"),
        ServiceName("apigatewayv2", "ApiGatewayV2"),
        ServiceName("appconfig", "AppConfig"),
        ServiceName("application-autoscaling", "ApplicationAutoScaling"),
        ServiceName("application-insights", "ApplicationInsights"),
        ServiceName("appmesh", "AppMesh"),
        ServiceName("appstream", "AppStream"),
        ServiceName("appsync", "AppSync"),
        ServiceName("athena", "Athena"),
        ServiceName("autoscaling", "AutoScaling"),
        ServiceName("autoscaling-plans", "AutoScalingPlans"),
        ServiceName("backup", "Backup"),
        ServiceName("batch", "Batch"),
        ServiceName("budgets", "Budgets"),
        ServiceName("ce", "CostExplorer"),
        ServiceName("chime", "Chime"),
        ServiceName("cloud9", "Cloud9"),
        ServiceName("clouddirectory", "CloudDirectory"),
        ServiceName("cloudformation", "CloudFormation"),
        ServiceName("cloudfront", "CloudFront"),
        ServiceName("cloudhsm", "CloudHSM"),
        ServiceName("cloudhsmv2", "CloudHSMV2"),
        ServiceName("cloudsearch", "CloudSearch"),
        ServiceName("cloudsearchdomain", "CloudSearchDomain"),
        ServiceName("cloudtrail", "CloudTrail"),
        ServiceName("cloudwatch", "CloudWatch"),
        ServiceName("codebuild", "CodeBuild"),
        ServiceName("codecommit", "CodeCommit"),
        ServiceName("codedeploy", "CodeDeploy"),
        ServiceName("codeguru-reviewer", "CodeGuruReviewer"),
        ServiceName("codeguruprofiler", "CodeGuruProfiler"),
        ServiceName("codepipeline", "CodePipeline"),
        ServiceName("codestar", "CodeStar"),
        ServiceName("codestar-notifications", "CodeStarNotifications"),
        ServiceName("cognito-identity", "CognitoIdentity"),
        ServiceName("cognito-idp", "CognitoIdentityProvider"),
        ServiceName("cognito-sync", "CognitoSync"),
        ServiceName("comprehend", "Comprehend"),
        ServiceName("comprehendmedical", "ComprehendMedical"),
        ServiceName("compute-optimizer", "ComputeOptimizer"),
        ServiceName("config", "ConfigService"),
        ServiceName("connect", "Connect"),
        ServiceName("connectparticipant", "ConnectParticipant"),
        ServiceName("cur", "CostandUsageReportService"),
        ServiceName("dataexchange", "DataExchange"),
        ServiceName("datapipeline", "DataPipeline"),
        ServiceName("datasync", "DataSync"),
        ServiceName("dax", "DAX"),
        ServiceName("devicefarm", "DeviceFarm"),
        ServiceName("directconnect", "DirectConnect"),
        ServiceName("discovery", "ApplicationDiscoveryService"),
        ServiceName("dlm", "DLM"),
        ServiceName("dms", "DatabaseMigrationService"),
        ServiceName("docdb", "DocDB"),
        ServiceName("ds", "DirectoryService"),
        ServiceName("dynamodb", "DynamoDB"),
        ServiceName("dynamodbstreams", "DynamoDBStreams"),
        ServiceName("ebs", "EBS"),
        ServiceName("ec2", "EC2"),
        ServiceName("ec2-instance-connect", "EC2InstanceConnect"),
        ServiceName("ecr", "ECR"),
        ServiceName("ecs", "ECS"),
        ServiceName("efs", "EFS"),
        ServiceName("eks", "EKS"),
        ServiceName("elastic-inference", "ElasticInference"),
        ServiceName("elasticache", "ElastiCache"),
        ServiceName("elasticbeanstalk", "ElasticBeanstalk"),
        ServiceName("elastictranscoder", "ElasticTranscoder"),
        ServiceName("elb", "ElasticLoadBalancing"),
        ServiceName("elbv2", "ElasticLoadBalancingv2"),
        ServiceName("emr", "EMR"),
        ServiceName("es", "ElasticsearchService"),
        ServiceName("events", "EventBridge"),
        ServiceName("firehose", "Firehose"),
        ServiceName("fms", "FMS"),
        ServiceName("forecast", "ForecastService"),
        ServiceName("forecastquery", "ForecastQueryService"),
        ServiceName("frauddetector", "FraudDetector"),
        ServiceName("fsx", "FSx"),
        ServiceName("gamelift", "GameLift"),
        ServiceName("glacier", "Glacier"),
        ServiceName("globalaccelerator", "GlobalAccelerator"),
        ServiceName("glue", "Glue"),
        ServiceName("greengrass", "Greengrass"),
        ServiceName("groundstation", "GroundStation"),
        ServiceName("guardduty", "GuardDuty"),
        ServiceName("health", "Health"),
        ServiceName("iam", "IAM"),
        ServiceName("imagebuilder", "Imagebuilder"),
        ServiceName("importexport", "ImportExport"),
        ServiceName("inspector", "Inspector"),
        ServiceName("iot", "IoT"),
        ServiceName("iot-data", "IoTDataPlane"),
        ServiceName("iot-jobs-data", "IoTJobsDataPlane"),
        ServiceName("iot1click-devices", "IoT1ClickDevicesService"),
        ServiceName("iot1click-projects", "IoT1ClickProjects"),
        ServiceName("iotanalytics", "IoTAnalytics"),
        ServiceName("iotevents", "IoTEvents"),
        ServiceName("iotevents-data", "IoTEventsData"),
        ServiceName("iotsecuretunneling", "IoTSecureTunneling"),
        ServiceName("iotthingsgraph", "IoTThingsGraph"),
        ServiceName("kafka", "Kafka"),
        ServiceName("kendra", "Kendra"),
        ServiceName("kinesis", "Kinesis"),
        ServiceName("kinesis-video-archived-media", "KinesisVideoArchivedMedia"),
        ServiceName("kinesis-video-media", "KinesisVideoMedia"),
        ServiceName("kinesis-video-signaling", "KinesisVideoSignalingChannels"),
        ServiceName("kinesisanalytics", "KinesisAnalytics"),
        ServiceName("kinesisanalyticsv2", "KinesisAnalyticsV2"),
        ServiceName("kinesisvideo", "KinesisVideo"),
        ServiceName("kms", "KMS"),
        ServiceName("lakeformation", "LakeFormation"),
        ServiceName("lambda", "Lambda"),
        ServiceName("lex-models", "LexModelBuildingService"),
        ServiceName("lex-runtime", "LexRuntimeService"),
        ServiceName("license-manager", "LicenseManager"),
        ServiceName("lightsail", "Lightsail"),
        ServiceName("logs", "CloudWatchLogs"),
        ServiceName("machinelearning", "MachineLearning"),
        ServiceName("macie", "Macie"),
        ServiceName("managedblockchain", "ManagedBlockchain"),
        ServiceName("marketplace-catalog", "MarketplaceCatalog"),
        ServiceName("marketplace-entitlement", "MarketplaceEntitlementService"),
        ServiceName("marketplacecommerceanalytics", "MarketplaceCommerceAnalytics"),
        ServiceName("mediaconnect", "MediaConnect"),
        ServiceName("mediaconvert", "MediaConvert"),
        ServiceName("medialive", "MediaLive"),
        ServiceName("mediapackage", "MediaPackage"),
        ServiceName("mediapackage-vod", "MediaPackageVod"),
        ServiceName("mediastore", "MediaStore"),
        ServiceName("mediastore-data", "MediaStoreData"),
        ServiceName("mediatailor", "MediaTailor"),
        ServiceName("meteringmarketplace", "MarketplaceMetering"),
        ServiceName("mgh", "MigrationHub"),
        ServiceName("migrationhub-config", "MigrationHubConfig"),
        ServiceName("mobile", "Mobile"),
        ServiceName("mq", "MQ"),
        ServiceName("mturk", "MTurk"),
        ServiceName("neptune", "Neptune"),
        ServiceName("networkmanager", "NetworkManager"),
        ServiceName("opsworks", "OpsWorks"),
        ServiceName("opsworkscm", "OpsWorksCM"),
        ServiceName("organizations", "Organizations"),
        ServiceName("outposts", "Outposts"),
        ServiceName("personalize", "Personalize"),
        ServiceName("personalize-events", "PersonalizeEvents"),
        ServiceName("personalize-runtime", "PersonalizeRuntime"),
        ServiceName("pi", "PI"),
        ServiceName("pinpoint", "Pinpoint"),
        ServiceName("pinpoint-email", "PinpointEmail"),
        ServiceName("pinpoint-sms-voice", "PinpointSMSVoice"),
        ServiceName("polly", "Polly"),
        ServiceName("pricing", "Pricing"),
        ServiceName("qldb", "QLDB"),
        ServiceName("qldb-session", "QLDBSession"),
        ServiceName("quicksight", "QuickSight"),
        ServiceName("ram", "RAM"),
        ServiceName("rds", "RDS"),
        ServiceName("rds-data", "RDSDataService"),
        ServiceName("redshift", "Redshift"),
        ServiceName("rekognition", "Rekognition"),
        ServiceName("resource-groups", "ResourceGroups"),
        ServiceName("resourcegroupstaggingapi", "ResourceGroupsTaggingAPI"),
        ServiceName("robomaker", "RoboMaker"),
        ServiceName("route53", "Route53"),
        ServiceName("route53domains", "Route53Domains"),
        ServiceName("route53resolver", "Route53Resolver"),
        ServiceName("s3", "S3"),
        ServiceName("s3control", "S3Control"),
        ServiceName("sagemaker", "SageMaker"),
        ServiceName("sagemaker-a2i-runtime", "AugmentedAIRuntime"),
        ServiceName("sagemaker-runtime", "SageMakerRuntime"),
        ServiceName("savingsplans", "SavingsPlans"),
        ServiceName("schemas", "Schemas"),
        ServiceName("sdb", "SimpleDB"),
        ServiceName("secretsmanager", "SecretsManager"),
        ServiceName("securityhub", "SecurityHub"),
        ServiceName("serverlessrepo", "ServerlessApplicationRepository"),
        ServiceName("service-quotas", "ServiceQuotas"),
        ServiceName("servicecatalog", "ServiceCatalog"),
        ServiceName("servicediscovery", "ServiceDiscovery"),
        ServiceName("ses", "SES"),
        ServiceName("sesv2", "SESV2"),
        ServiceName("shield", "Shield"),
        ServiceName("signer", "Signer"),
        ServiceName("sms", "SMS"),
        ServiceName("sms-voice", "SMSVoice"),
        ServiceName("snowball", "Snowball"),
        ServiceName("sns", "SNS"),
        ServiceName("sqs", "SQS"),
        ServiceName("ssm", "SSM"),
        ServiceName("sso", "SSO"),
        ServiceName("sso-oidc", "SSOOIDC"),
        ServiceName("stepfunctions", "SFN"),
        ServiceName("storagegateway", "StorageGateway"),
        ServiceName("sts", "STS"),
        ServiceName("support", "Support"),
        ServiceName("swf", "SWF"),
        ServiceName("textract", "Textract"),
        ServiceName("transcribe", "TranscribeService"),
        ServiceName("transfer", "Transfer"),
        ServiceName("translate", "Translate"),
        ServiceName("waf", "WAF"),
        ServiceName("waf-regional", "WAFRegional"),
        ServiceName("wafv2", "WAFV2"),
        ServiceName("workdocs", "WorkDocs"),
        ServiceName("worklink", "WorkLink"),
        ServiceName("workmail", "WorkMail"),
        ServiceName("workmailmessageflow", "WorkMailMessageFlow"),
        ServiceName("workspaces", "WorkSpaces"),
        ServiceName("xray", "XRay"),
    )
    ITEM_MAP = {i.name: i for i in ITEMS}

    ec2 = ITEM_MAP["ec2"]
    iam = ITEM_MAP["iam"]
    s3 = ITEM_MAP["s3"]
    cloudwatch = ITEM_MAP["cloudwatch"]
    opsworks = ITEM_MAP["opsworks"]
    sns = ITEM_MAP["sns"]
    glacier = ITEM_MAP["glacier"]
    dynamodb = ITEM_MAP["dynamodb"]
    sqs = ITEM_MAP["sqs"]
    cloudformation = ITEM_MAP["cloudformation"]
    cloudsearchdomain = ITEM_MAP["cloudsearchdomain"]
    logs = ITEM_MAP["logs"]

    @classmethod
    def find(cls, name: str) -> ServiceName:
        try:
            return cls.ITEM_MAP[name]
        except KeyError:
            raise ValueError(f"Unknown service {name}")



"""
Processors for parsing and writing modules.
"""
from pathlib import Path

from boto3.session import Session

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.version import __version__ as version
from mypy_boto3_builder.parsers.master_package import parse_master_package
from mypy_boto3_builder.parsers.service_package import parse_service_package
from mypy_boto3_builder.structures.boto3_stubs_package import Boto3StubsPackage
from mypy_boto3_builder.structures.master_package import MasterPackage
from mypy_boto3_builder.structures.service_package import ServicePackage
from mypy_boto3_builder.utils.nice_path import NicePath
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.writers.boto3_stubs_package import write_boto3_stubs_package
from mypy_boto3_builder.writers.master_package import write_master_package
from mypy_boto3_builder.writers.service_package import write_service_package


logger = get_logger()


def process_boto3_stubs(output_path: Path) -> Boto3StubsPackage:
    """
    Parse and write stubs package `boto3_stubs`.

    Arguments:
        output_path -- Package output path.

    Return:
        Parsed Boto3StubsPackage.
    """
    logger.debug(f"Parsing boto3 stubs")
    boto3_stub_package = Boto3StubsPackage()
    logger.debug(f"Writing boto3 stubs to {NicePath(output_path)}")

    modified_paths = write_boto3_stubs_package(boto3_stub_package, output_path)
    for modified_path in modified_paths:
        logger.debug(f"Updated {NicePath(modified_path)}")

    return boto3_stub_package


def process_master(session: Session, output_path: Path) -> MasterPackage:
    """
    Parse and write master package `mypy_boto3`.

    Arguments:
        session -- boto3 session.
        output_path -- Package output path.

    Return:
        Parsed MasterPackage.
    """
    logger.debug(f"Parsing master")
    master_package = parse_master_package(session)
    logger.debug(f"Writing master to {NicePath(output_path)}")

    modified_paths = write_master_package(master_package, output_path)
    for modified_path in modified_paths:
        logger.debug(f"Updated {NicePath(modified_path)}")

    return master_package


def process_service(
    session: Session, service_name: ServiceName, output_path: Path
) -> ServicePackage:
    """
    Parse and write service package `mypy_boto3_*`.

    Arguments:
        session -- boto3 session.
        service_name -- Target service name.
        output_path -- Package output path.

    Return:
        Parsed ServicePackage.
    """
    logger.debug(f"Parsing {service_name.boto3_name}")
    service_module = parse_service_package(session, service_name)
    logger.debug(f"Writing {service_name.boto3_name} to {NicePath(output_path)}")

    modified_paths = write_service_package(service_module, output_path)
    module_file_names = [i.value for i in ServiceModuleName]
    for modified_path in modified_paths:
        if modified_path.stem in module_file_names:
            logger.info(f"Updated {NicePath(modified_path)}")
        else:
            logger.debug(f"Updated {NicePath(modified_path)}")

    return service_module



"""
Wrapper for path that represents it as relative to workdir.
"""
from pathlib import Path


class NicePath:
    """
    Wrapper for path that represents it as relative to workdir.

    Arguments:
        path -- Original path.
    """

    def __init__(self, path: Path) -> None:
        self.path = path

    def __str__(self) -> str:
        path = self.path
        try:
            path = path.relative_to(Path.cwd())
        except ValueError:
            pass

        return path.as_posix()



# pylint: disable=unused-argument,multiple-statements,super-init-not-called
from typing import Iterable, Any

import botocore.exceptions

class Boto3Error(Exception): ...
class ResourceLoadException(Boto3Error): ...
class NoVersionFound(Boto3Error): ...

class UnknownAPIVersionError(Boto3Error, botocore.exceptions.DataNotFoundError):
    def __init__(
        self,
        service_name: str,
        bad_api_version: str,
        available_api_versions: Iterable[str],
    ): ...

class ResourceNotExistsError(Boto3Error, botocore.exceptions.DataNotFoundError):
    def __init__(
        self,
        service_name: str,
        available_services: Iterable[str],
        has_low_level_client: bool,
    ) -> None: ...

class RetriesExceededError(Boto3Error):
    def __init__(
        self, last_exception: Boto3Error, msg: str = "Max Retries Exceeded"
    ) -> None: ...

class S3TransferFailedError(Boto3Error):
    pass

class S3UploadFailedError(Boto3Error):
    pass

class DynamoDBOperationNotSupportedError(Boto3Error):
    def __init__(self, operation: str, value: Any) -> None: ...

DynanmoDBOperationNotSupportedError = DynamoDBOperationNotSupportedError

class DynamoDBNeedsConditionError(Boto3Error):
    def __init__(self, value: Any) -> None: ...

class DynamoDBNeedsKeyConditionError(Boto3Error): ...



"""
Structure for boto3-stubs module.
"""

from dataclasses import dataclass, field
from typing import List, Tuple

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.constants import PYPI_NAME, MODULE_NAME
from mypy_boto3_builder.structures.package import Package
from mypy_boto3_builder.structures.service_package import ServicePackage


@dataclass
class MasterPackage(Package):
    """
    Structure for mypy-boto3 package.
    """

    name: str = MODULE_NAME
    pypi_name: str = PYPI_NAME
    service_packages: List[ServicePackage] = field(default_factory=lambda: [])

    @property
    def essential_service_names(self) -> List[ServiceName]:
        result: List[ServiceName] = []
        for service_name in self.service_names:
            if service_name.is_essential():
                result.append(service_name)
        return result

    @property
    def service_names(self) -> Tuple[ServiceName, ...]:
        return ServiceNameCatalog.ITEMS



"""
Wrapper for simple type annotations from this module.
"""
from __future__ import annotations

from typing import Optional

from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.internal_import_record import (
    InternalImportRecord,
)
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName


class InternalImport(FakeAnnotation):
    """
    Wrapper for simple type annotations from this module.

    Arguments:
        name -- Import name.
        service_name -- Service that import belongs to.
        module_name -- Service module name.
    """

    def __init__(
        self,
        name: str,
        service_name: Optional[ServiceName] = None,
        module_name: ServiceModuleName = ServiceModuleName.service_resource,
    ) -> None:
        self.name = name
        self.service_name = service_name
        self.module_name = module_name

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        return f"{self.scope}.{self.name}"

    @property
    def scope(self) -> str:
        return f"{self.module_name.name}_scope"

    def get_import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        if self.service_name is not None:
            return ImportRecord(
                source=ImportString(
                    self.service_name.module_name, self.module_name.name
                ),
                alias=self.scope,
            )
        return InternalImportRecord(self.module_name, alias=self.scope)

    def copy(self) -> InternalImport:
        """
        Create a copy of type annotation wrapper.
        """
        return InternalImport(self.name, self.service_name, self.module_name)



from typing import Dict, List, Any

from boto3.session import Session
from botocore.exceptions import UnknownServiceError
from botocore import xform_name
from botocore.session import Session as BotocoreSession
from botocore.model import (
    Shape,
    OperationModel,
    ServiceModel,
    StructureShape,
    MapShape,
    ListShape,
    StringShape,
)

from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.type_maps.operation_type_map import get_argument_type_stub
from mypy_boto3_builder.type_maps.typed_dicts import (
    waiter_config_type,
    paginator_config_type,
)


class ShapeParserError(Exception):
    pass


class ShapeParser:
    SHAPE_TYPE_MAP = {
        "integer": Type.int,
        "long": Type.int,
        "boolean": Type.bool,
        "double": Type.float,
        "float": Type.float,
        "timestamp": ExternalImport(ImportString("datetime"), "datetime"),
        "blob": TypeSubscript(Type.Union, [Type.bytes, Type.IO]),
    }

    ARGUMENT_ALIASES: Dict[str, Dict[str, Dict[str, str]]] = {
        ServiceNameCatalog.cloudsearchdomain.boto3_name: {
            "Search": {"return": "returnFields"}
        },
        ServiceNameCatalog.logs.boto3_name: {"CreateExportTask": {"from": "fromTime"}},
        ServiceNameCatalog.ec2.boto3_name: {"*": {"Filter": "Filters"}},
        ServiceNameCatalog.s3.boto3_name: {
            "PutBucketAcl": {"ContentMD5": "None"},
            "PutBucketCors": {"ContentMD5": "None"},
            "PutBucketLifecycle": {"ContentMD5": "None"},
            "PutBucketLogging": {"ContentMD5": "None"},
            "PutBucketNotification": {"ContentMD5": "None"},
            "PutBucketPolicy": {"ContentMD5": "None"},
            "PutBucketReplication": {"ContentMD5": "None"},
            "PutBucketRequestPayment": {"ContentMD5": "None"},
            "PutBucketTagging": {"ContentMD5": "None"},
            "PutBucketVersioning": {"ContentMD5": "None"},
            "PutBucketWebsite": {"ContentMD5": "None"},
            "PutObjectAcl": {"ContentMD5": "None"},
        },
    }

    def __init__(self, session: Session, service_name: ServiceName):
        loader = session._loader  # pylint: disable=protected-access
        botocore_session: BotocoreSession = session._session  # pylint: disable=protected-access
        service_data = botocore_session.get_service_data(service_name.boto3_name)
        self.service_name = service_name
        self.service_model = ServiceModel(service_data, service_name.boto3_name)
        self._typed_dict_map: Dict[str, TypeTypedDict] = {}
        self._waiters_shape: Shape = {}
        try:
            self._waiters_shape = loader.load_service_model(
                service_name.boto3_name, "waiters-2"
            )
        except UnknownServiceError:
            pass
        self._paginators_shape: Shape = {}
        try:
            self._paginators_shape = loader.load_service_model(
                service_name.boto3_name, "paginators-1"
            )
        except UnknownServiceError:
            pass
        self._resources_shape: Shape = {}
        try:
            self._resources_shape = loader.load_service_model(
                service_name.boto3_name, "resources-1"
            )
        except UnknownServiceError:
            pass

        self.logger = get_logger()

    def _get_shape(self, name: str) -> Shape:
        return self.service_model.shape_for(name)

    def _get_operation(self, name: str) -> OperationModel:
        return self.service_model.operation_model(name)

    def _get_operation_names(self) -> List[str]:
        return list(
            self.service_model.operation_names
        )  # pylint: disable=not-an-iterable

    @staticmethod
    def _rename_input_member(
        operation: OperationModel, old_name: str, new_name: str
    ) -> None:
        members = operation.input_shape.members
        members[new_name] = members[old_name]
        del members[old_name]

    @staticmethod
    def _delete_input_member(operation: OperationModel, name: str) -> None:
        members = operation.input_shape.members
        del members[name]

    def _get_paginator(self, name: str) -> Shape:
        try:
            return self._paginators_shape["pagination"][name]
        except KeyError:
            raise ShapeParserError(f"Unknown paginator: {name}")

    def _get_service_resource(self) -> Shape:
        return self._resources_shape["service"]

    def _get_resource_shape(self, name: str) -> Shape:
        try:
            return self._resources_shape["resources"][name]
        except KeyError:
            raise ShapeParserError(f"Unknown resource: {name}")

    def _get_resource_collection_shape(self, resouce_name: str, name: str) -> Shape:
        try:
            return self._get_resource_shape(resouce_name)["hasMany"][name]
        except KeyError:
            raise ShapeParserError(
                f"Unknown resource collection: {resouce_name}.{name}"
            )

    def _get_resource_action_shape(self, resouce_name: str, name: str) -> Shape:
        try:
            return self._get_resource_shape(resouce_name)["actions"][name]
        except KeyError:
            raise ShapeParserError(f"Unknown resource action: {resouce_name}.{name}")

    def get_paginator_names(self) -> List[str]:
        result: List[str] = []
        for name in self._paginators_shape.get("pagination", []):
            result.append(name)
        result.sort()
        return result

    def _get_argument_alias(self, operation_name: str, argument_name: str) -> str:
        service_map = self.ARGUMENT_ALIASES.get(self.service_name.boto3_name)
        if not service_map:
            return argument_name

        operation_map: Dict[str, str] = {}
        if "*" in service_map:
            operation_map = service_map["*"]
        if operation_name in service_map:
            operation_map = service_map[operation_name]

        if not operation_map:
            return argument_name

        if argument_name not in operation_map:
            return argument_name

        return operation_map[argument_name]

    def _parse_arguments(
        self, operation_name: str, shape: StructureShape
    ) -> List[Argument]:
        result: List[Argument] = []
        required = shape.required_members
        for argument_name, argument_shape in shape.members.items():
            argument_alias = self._get_argument_alias(operation_name, argument_name)
            if argument_alias == "None":
                continue
            argument_type_stub = get_argument_type_stub(operation_name, argument_name)
            if argument_type_stub is not None:
                argument_type = argument_type_stub
            else:
                argument_type = self._parse_shape(argument_shape)
            argument = Argument(argument_alias, argument_type)
            if argument_name not in required:
                argument.default = Type.none
            result.append(argument)

        result.sort(key=lambda x: not x.required)
        return result

    def _parse_return_type(self, shape: Shape) -> FakeAnnotation:
        return self._parse_shape(shape)

    def get_client_method_map(self) -> Dict[str, Method]:
        result: Dict[str, Method] = {
            "can_paginate": Method(
                "can_paginate",
                [Argument("self", None), Argument("operation_name", Type.str)],
                Type.bool,
            )
        }
        for operation_name in self._get_operation_names():
            operation_model = self._get_operation(operation_name)
            return_type: FakeAnnotation = Type.none
            arguments: List[Argument] = [Argument("self", None)]

            if operation_model.input_shape is not None:
                arguments.extend(
                    self._parse_arguments(operation_name, operation_model.input_shape)
                )

            if operation_model.output_shape is not None:
                return_type = self._parse_return_type(operation_model.output_shape)

            method_name = xform_name(operation_name)
            method = Method(
                name=method_name, arguments=arguments, return_type=return_type
            )
            result[method.name] = method

        return result

    @staticmethod
    def _parse_shape_string(shape: StringShape) -> FakeAnnotation:
        if not shape.enum:
            return Type.str

        type_literal = TypeLiteral()
        for option in shape.enum:
            type_literal.add_literal_child(option)

        return type_literal

    def _parse_shape_map(self, shape: MapShape) -> FakeAnnotation:
        type_subscript = TypeSubscript(Type.Dict)
        if shape.key:
            type_subscript.add_child(self._parse_shape(shape.key))
        else:
            type_subscript.add_child(Type.str)
        if shape.value:
            type_subscript.add_child(self._parse_shape(shape.value))
        else:
            type_subscript.add_child(Type.Any)
        return type_subscript

    def _parse_shape_structure(self, shape: StructureShape) -> FakeAnnotation:
        if not shape.members.items():
            return TypeSubscript(Type.Dict, [Type.str, Type.Any])

        required = shape.required_members
        typed_dict_name = f"{shape.name}TypeDef"
        if typed_dict_name in self._typed_dict_map:
            return self._typed_dict_map[typed_dict_name]
        typed_dict = TypeTypedDict(typed_dict_name)
        self._typed_dict_map[typed_dict_name] = typed_dict
        for attr_name, attr_shape in shape.members.items():
            typed_dict.add_attribute(
                attr_name, self._parse_shape(attr_shape), attr_name in required,
            )
        return typed_dict

    def _parse_shape_list(self, shape: ListShape) -> FakeAnnotation:
        type_subscript = TypeSubscript(Type.List)
        if shape.member:
            type_subscript.add_child(self._parse_shape(shape.member))
        else:
            type_subscript.add_child(Type.Any)
        return type_subscript

    def _parse_shape(self, shape: Shape) -> FakeAnnotation:
        if shape.type_name in self.SHAPE_TYPE_MAP:
            return self.SHAPE_TYPE_MAP[shape.type_name]

        if isinstance(shape, StringShape):
            return self._parse_shape_string(shape)

        if isinstance(shape, MapShape):
            return self._parse_shape_map(shape)

        if isinstance(shape, StructureShape):
            return self._parse_shape_structure(shape)

        if isinstance(shape, ListShape):
            return self._parse_shape_list(shape)

        if shape.type_name in self._resources_shape["resources"]:
            return InternalImport(shape.type_name)

        self.logger.warning(f"Unknown shape: {shape}")
        return Type.Any

    def get_paginate_method(self, paginator_name: str) -> Method:
        operation_name = paginator_name
        paginator_shape = self._get_paginator(paginator_name)
        operation_shape = self._get_operation(operation_name)
        skip_argument_names: List[str] = []
        input_token = paginator_shape["input_token"]
        if isinstance(input_token, list):
            skip_argument_names.extend(input_token)
        else:
            skip_argument_names.append(input_token)
        if "limit_key" in paginator_shape:
            skip_argument_names.append(paginator_shape["limit_key"])

        arguments: List[Argument] = [Argument("self", None)]

        if operation_shape.input_shape is not None:
            for argument in self._parse_arguments(
                operation_name, operation_shape.input_shape
            ):
                if argument.name in skip_argument_names:
                    continue
                arguments.append(argument)

        arguments.append(Argument("PaginationConfig", paginator_config_type, Type.none))

        return_type: FakeAnnotation = Type.none
        if operation_shape.output_shape is not None:
            return_type = self._parse_return_type(operation_shape.output_shape)

        return Method("paginate", arguments, return_type)

    def get_wait_method(self, waiter_name: str) -> Method:
        operation_name = self._waiters_shape["waiters"][waiter_name]["operation"]
        operation_shape = self._get_operation(operation_name)

        arguments: List[Argument] = [Argument("self", None)]

        if operation_shape.input_shape is not None:
            arguments.extend(
                self._parse_arguments(operation_name, operation_shape.input_shape)
            )

        arguments.append(Argument("WaiterConfig", waiter_config_type, Type.none))

        return Method(name="wait", arguments=arguments, return_type=Type.none)

    def get_service_resource_method_map(self) -> Dict[str, Method]:
        result: Dict[str, Method] = {}
        service_resource_shape = self._get_service_resource()
        for action_name, action_shape in service_resource_shape.get(
            "actions", {}
        ).items():
            method = self._get_resource_method(action_name, action_shape)
            result[method.name] = method

        return result

    def get_resource_method_map(self, resource_name: str) -> Dict[str, Method]:
        resource_shape = self._get_resource_shape(resource_name)
        result: Dict[str, Method] = {
            "get_available_subresources": Method(
                "get_available_subresources",
                [Argument("self", None)],
                TypeSubscript(Type.List, [Type.str]),
            ),
            "load": Method("load", [Argument("self", None)], Type.none),
            "reload": Method("reload", [Argument("self", None)], Type.none),
        }

        for action_name, action_shape in resource_shape.get("actions", {}).items():
            method = self._get_resource_method(action_name, action_shape)
            result[method.name] = method

        for waiter_name in resource_shape.get("waiters", {}):
            method = Method(
                f"wait_until_{xform_name(waiter_name)}",
                [Argument("self", None)],
                Type.none,
            )
            result[method.name] = method

        return result

    def _get_resource_method(
        self, action_name: str, action_shape: Dict[str, Any]
    ) -> Method:
        return_type: FakeAnnotation = Type.none
        arguments: List[Argument] = [Argument("self", None)]
        if "resource" in action_shape:
            return_type = self._parse_return_type(
                Shape("resource", action_shape["resource"])
            )

        if "request" in action_shape:
            operation_name = action_shape["request"]["operation"]
            operation_shape = self._get_operation(operation_name)
            skip_argument_names: List[str] = [
                i["target"]
                for i in action_shape["request"].get("params", {})
                if i["source"] == "identifier"
            ]
            if operation_shape.input_shape is not None:
                for argument in self._parse_arguments(
                    operation_name, operation_shape.input_shape
                ):
                    if argument.name not in skip_argument_names:
                        arguments.append(argument)
            if operation_shape.output_shape is not None:
                return_type = self._parse_shape(operation_shape.output_shape)

        return Method(
            name=xform_name(action_name), arguments=arguments, return_type=return_type
        )



"""
Wrapper for `typing/typing_extensions.TypedDict` type annotations.
"""
from __future__ import annotations

from typing import Iterable, Set, List

from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.internal_import_record import (
    InternalImportRecord,
)
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class TypedDictAttribute:
    """
    TypedDict attribute wrapper.

    Arguments:
        name -- Attribute name.
        type_annotation -- Attribute type annotation.
        required -- Whether the attribute has to be set.
    """

    def __init__(self, name: str, type_annotation: FakeAnnotation, required: bool):
        self.name = name
        self.type_annotation = type_annotation
        self.required = required

    def render(self) -> str:
        """
        Render attribute to use in class-based TypedDict definition.

        Returns:
            A string with argument definition.
        """
        return f"{self.name}: {self.type_annotation.render()}"


class TypeTypedDict(FakeAnnotation):
    """
    Wrapper for `typing/typing_extensions.TypedDict` type annotations.

    Arguments:
        name -- Type name.
        children -- Typed dict attributes.
        docstring -- Docstring for render.
    """

    def __init__(
        self,
        name: str,
        children: Iterable[TypedDictAttribute] = (),
        docstring: str = "",
    ) -> None:
        self.name = name
        self.children = list(children)
        self.docstring = docstring

    def get_attribute(self, name: str) -> TypedDictAttribute:
        for child in self.children:
            if child.name == name:
                return child

        raise ValueError(f"No child with name {name}")

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        if parent_name and parent_name == self.name:
            return f'"{self.name}"'

        return self.name

    def get_import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """
        return InternalImportRecord(ServiceModuleName.type_defs, name=self.name)

    def get_types(self) -> Set[FakeAnnotation]:
        """
        Get set with itself.

        To get child types, `get_children_types` has to be used.

        Returns:
            A set of type annotations.
        """
        return {self}

    def add_attribute(
        self, name: str, type_annotation: FakeAnnotation, required: bool
    ) -> None:
        """
        Add new attribute to a dictionary.

        Arguments:
            name -- Argument name.
            type_annotation -- Argument type annotation.
            required -- Whether argument has to be set.
        """
        self.children.append(TypedDictAttribute(name, type_annotation, required))

    def is_dict(self) -> bool:
        """
        Always True as it is a TypedDict.
        """
        return True

    def is_typed_dict(self) -> bool:
        """
        Always True as it is a TypedDict.
        """
        return True

    def render_class(self) -> str:
        """
        Render class-based definition for debugging.
        """
        children = "\n".join([f"     {i.render()}" for i in self.children])
        return f"class {self.name}:\n{children}"

    def has_optional(self) -> bool:
        """
        Whether TypedDict has optional keys.
        """
        for child in self.children:
            if not child.required:
                return True
        return False

    def has_required(self) -> bool:
        """
        Whether TypedDict has required keys.
        """
        for child in self.children:
            if child.required:
                return True
        return False

    def has_both(self) -> bool:
        """
        Whether TypedDict has both optional and required keys.
        """
        return self.has_required() and self.has_optional()

    def get_required(self) -> List[TypedDictAttribute]:
        result: List[TypedDictAttribute] = []
        for child in self.children:
            if child.required:
                result.append(child)
        return result

    def get_optional(self) -> List[TypedDictAttribute]:
        result: List[TypedDictAttribute] = []
        for child in self.children:
            if not child.required:
                result.append(child)
        return result

    def copy(self) -> TypeTypedDict:
        """
        Create a copy of type annotation wrapper.
        """
        return TypeTypedDict(self.name, list(self.children), docstring=self.docstring)

    def is_same(self, other: TypeTypedDict) -> bool:
        children = [i.render() for i in self.children]
        other_children = [i.render() for i in other.children]
        return other_children == children

    def get_children_types(self) -> Set[FakeAnnotation]:
        result: Set[FakeAnnotation] = set()
        for child in self.children:
            result.update(child.type_annotation.get_types())
        return result



from __future__ import annotations

from dataclasses import dataclass, field
from typing import List, Set, Optional, Iterable, Dict

from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.import_helpers.import_record_group import ImportRecordGroup
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_typed_dict import TypeTypedDict
from mypy_boto3_builder.structures.service_resource import ServiceResource
from mypy_boto3_builder.structures.waiter import Waiter
from mypy_boto3_builder.structures.paginator import Paginator
from mypy_boto3_builder.structures.client import Client
from mypy_boto3_builder.structures.function import Function
from mypy_boto3_builder.structures.package import Package


@dataclass
class ServicePackage(Package):
    service_name: ServiceName
    client: Client
    service_resource: Optional[ServiceResource] = None
    waiters: List[Waiter] = field(default_factory=lambda: [])
    paginators: List[Paginator] = field(default_factory=lambda: [])
    typed_dicts: List[TypeTypedDict] = field(default_factory=lambda: [])
    helper_functions: List[Function] = field(default_factory=lambda: [])

    def extract_typed_dicts(
        self,
        type_annotations: Iterable[FakeAnnotation],
        added: Dict[str, TypeTypedDict],
    ) -> List[TypeTypedDict]:
        result: List[TypeTypedDict] = []
        for type_annotation in sorted(type_annotations):
            if not isinstance(type_annotation, TypeTypedDict):
                continue
            if type_annotation.name in added:
                try:
                    assert added[type_annotation.name].is_same(type_annotation)
                except AssertionError:
                    print(type_annotation.render_class())
                    print(added[type_annotation.name].render_class())
                    raise ValueError(type_annotation.name)
                continue

            added[type_annotation.name] = type_annotation
            result.extend(
                self.extract_typed_dicts(type_annotation.get_children_types(), added)
            )
            result.append(type_annotation)
        return result

    def get_types(self) -> Set[FakeAnnotation]:
        types: Set[FakeAnnotation] = set()
        types.update(self.client.get_types())
        if self.service_resource:
            types.update(self.service_resource.get_types())
        for waiter in self.waiters:
            types.update(waiter.get_types())
        for paginator in self.paginators:
            types.update(paginator.get_types())
        return types

    def get_init_import_record_groups(self) -> List[ImportRecordGroup]:
        import_records: Set[ImportRecord] = set()
        import_records.add(
            ImportRecord(
                ImportString(
                    self.service_name.module_name, ServiceModuleName.client.name
                ),
                self.client.name,
            )
        )
        if self.service_resource:
            import_records.add(
                ImportRecord(
                    ImportString(
                        self.service_name.module_name,
                        ServiceModuleName.service_resource.name,
                    ),
                    self.service_resource.name,
                )
            )
        for waiter in self.waiters:
            import_records.add(
                ImportRecord(
                    ImportString(
                        self.service_name.module_name, ServiceModuleName.waiter.name,
                    ),
                    waiter.name,
                )
            )
        for paginator in self.paginators:
            import_records.add(
                ImportRecord(
                    ImportString(
                        self.service_name.module_name, ServiceModuleName.paginator.name,
                    ),
                    paginator.name,
                )
            )

        return ImportRecordGroup.from_import_records(import_records)

    def get_init_all_names(self) -> List[str]:
        result = [self.client.name]
        if self.service_resource:
            result.append(self.service_resource.name)
        for waiter in self.waiters:
            result.append(waiter.name)
        for paginator in self.paginators:
            result.append(paginator.name)
        return result

    def get_client_required_import_record_groups(self) -> List[ImportRecordGroup]:
        import_records: Set[ImportRecord] = set()
        for import_record in self.client.get_required_import_records():
            import_records.add(
                import_record.get_external(self.service_name.module_name)
            )
        for import_record in self.client.exceptions_class.get_required_import_records():
            import_records.add(
                import_record.get_external(self.service_name.module_name)
            )

        return ImportRecordGroup.from_import_records(import_records)

    def get_service_resource_required_import_record_groups(
        self,
    ) -> List[ImportRecordGroup]:
        if self.service_resource is None:
            return []

        class_import_records = self.service_resource.get_required_import_records()
        import_records: Set[ImportRecord] = set()
        for import_record in class_import_records:
            import_records.add(
                import_record.get_external(self.service_name.module_name)
            )
        return ImportRecordGroup.from_import_records(import_records)

    def get_paginator_required_import_record_groups(self) -> List[ImportRecordGroup]:
        import_records: Set[ImportRecord] = set()
        for paginator in self.paginators:
            for import_record in paginator.get_required_import_records():
                import_records.add(
                    import_record.get_external(self.service_name.module_name)
                )

        return ImportRecordGroup.from_import_records(import_records)

    def get_waiter_required_import_record_groups(self) -> List[ImportRecordGroup]:
        import_records: Set[ImportRecord] = set()
        for waiter in self.waiters:
            for import_record in waiter.get_required_import_records():
                import_records.add(
                    import_record.get_external(self.service_name.module_name)
                )

        return ImportRecordGroup.from_import_records(import_records)

    def get_type_defs_required_import_record_groups(self) -> List[ImportRecordGroup]:
        import_records: Set[ImportRecord] = set()
        if self.typed_dicts:
            import_records.add(
                ImportRecord(
                    ImportString("typing"),
                    "TypedDict",
                    min_version=(3, 8),
                    fallback=ImportRecord(
                        ImportString("typing_extensions"), "TypedDict"
                    ),
                )
            )
            for types_dict in self.typed_dicts:
                for type_annotation in types_dict.get_children_types():
                    import_record = type_annotation.get_import_record()
                    if import_record.is_type_defs():
                        continue
                    import_records.add(import_record)

        return ImportRecordGroup.from_import_records(import_records)

    def get_helpers_import_record_groups(self) -> List[ImportRecordGroup]:
        import_records: Set[ImportRecord] = set()
        import_records.add(ImportRecord(ImportString("boto3")))
        import_records.add(ImportRecord(ImportString("typing"), "Dict"))
        import_records.add(ImportRecord(ImportString("typing"), "Any"))
        for helper_function in self.helper_functions:
            for type_annotation in helper_function.get_types():
                import_record = type_annotation.get_import_record()
                import_records.add(import_record)

        return ImportRecordGroup.from_import_records(import_records)



"""
Logging utils.
"""
import logging
from typing import Any


__all__ = ("get_logger",)


class Logger(logging.Logger):
    """
    Standard Logger with optional panic mode.
    """

    # Set to true to raise RuntimeError on warning
    panic = False

    def warning(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        """
        Overriden default warning with no changes.

        Raises:
            RuntimeError -- If panic mode is on.
        """
        super().warning(msg, *args, **kwargs)
        if self.panic:
            raise RuntimeError(msg)

    def set_level(self, level: int) -> None:
        """
        Set level of logger and all handlers.

        Arguments:
            level -- Logging level.
        """
        self.setLevel(level)
        for handler in self.handlers:
            handler.setLevel(level)


logger = Logger("mypy_boto3_builder")


def get_logger(verbose: bool = False, panic: bool = False) -> Logger:
    """
    Get Logger instance.

    Arguments:
        verbose -- Set log level to DEBUG.
        panic -- Raise RuntimeError on warning.

    Returns:
        Overriden Logger.
    """
    if not logger.handlers:
        stream_handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s %(name)s: %(levelname)-8s %(message)s", datefmt="%H:%M:%S"
        )
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
        logger.set_level(logging.INFO)

    if verbose:
        logger.set_level(logging.DEBUG)

    if panic:
        logger.panic = True

    return logger



"""
Grouped by `source` import records for nicer rendering.
"""
from __future__ import annotations

from typing import Iterable, List, Set

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.import_helpers.import_record import ImportRecord


class ImportRecordGroup:
    """
    Grouped by `source` import records for nicer rendering.

    Arguments:
        source -- Common source for import records.
        import_records -- Grouped import records.
    """

    def __init__(
        self, source: ImportString, import_records: Iterable[ImportRecord]
    ) -> None:
        self.source = source
        self.import_records = list(import_records)

    @classmethod
    def from_import_records(
        cls, import_records: Iterable[ImportRecord]
    ) -> List[ImportRecordGroup]:
        """
        Get groups from `ImportRecord` list.

        Arguments:
            import_records -- Import records list.

        Returns:
            A list of generated `ImportRecordGroup`.
        """
        result: List[ImportRecordGroup] = []
        all_import_records: Set[ImportRecord] = set(import_records)

        for import_record in import_records:
            if import_record.fallback:
                all_import_records.add(ImportRecord(ImportString("sys")))

        for import_record in sorted(all_import_records):
            if not import_record:
                continue
            if import_record.is_builtins():
                continue
            if (
                not result
                or result[-1].source != import_record.source
                or not result[-1].import_records[0].name
                or import_record.is_standalone()
                or result[-1].import_records[0].is_standalone()
            ):
                result.append(ImportRecordGroup(import_record.source, [import_record]))
            else:
                result[-1].import_records.append(import_record)

        return result

    def is_builtins(self) -> bool:
        return self.source.startswith(ImportRecord.builtins_import_string)



# pylint: disable=unused-argument,multiple-statements,no-self-use
from __future__ import annotations

from typing import Any, List, Optional, Dict, Type
from types import TracebackType
import logging

from botocore.client import BaseClient

logger: logging.Logger

def register_table_methods(base_classes: List[type], **kwargs: Any) -> None: ...

class TableResource:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def batch_writer(
        self, overwrite_by_pkeys: Optional[List[str]] = None
    ) -> BatchWriter: ...

class BatchWriter:
    def __init__(
        self,
        table_name: str,
        client: BaseClient,
        flush_amount: int = 25,
        overwrite_by_pkeys: Optional[List[str]] = None,
    ) -> None: ...
    def put_item(self, Item: Dict[str, Any]) -> None: ...
    def delete_item(self, Key: Dict[str, Any]) -> None: ...
    def __enter__(self) -> BatchWriter: ...
    def __exit__(
        self, exc_type: Type[BaseException], exc_value: BaseException, tb: TracebackType
    ) -> None: ...



"""
Predefined FakeAnnotation instances.
"""
from typing import Union, Optional, Any, Dict, List, Callable, IO, overload

from mypy_boto3_builder.type_annotations.type_annotation import TypeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_class import TypeClass


class Type:
    """
    Predefined FakeAnnotation instances.
    """

    Union = TypeAnnotation(Union)
    Any = TypeAnnotation(Any)
    Dict = TypeAnnotation(Dict)
    List = TypeAnnotation(List)
    Optional = TypeAnnotation(Optional)
    Callable = TypeAnnotation(Callable)
    IO = TypeAnnotation(IO)
    overload = TypeAnnotation(overload)
    classmethod = TypeClass(classmethod)
    staticmethod = TypeClass(staticmethod)
    none = TypeConstant(None)
    str = TypeClass(str)
    bool = TypeClass(bool)
    bytes = TypeClass(bytes)
    int = TypeClass(int)
    float = TypeClass(float)
    Ellipsis = TypeConstant(...)



"""
Parser that produces `structures.ServiceModule`.
"""

from boto3.session import Session
from botocore import xform_name

from mypy_boto3_builder.structures.waiter import Waiter
from mypy_boto3_builder.structures.paginator import Paginator
from mypy_boto3_builder.structures.service_package import ServicePackage
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.parsers.client import parse_client
from mypy_boto3_builder.parsers.service_resource import parse_service_resource
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.logger import get_logger


def parse_service_package(
    session: Session, service_name: ServiceName
) -> ServicePackage:
    """
    Extract all data from boto3 service package.

    Arguments:
        session -- boto3 session.
        service_name -- Target service name.

    Returns:
        ServiceModule structure.
    """
    logger = get_logger()
    logger.debug("Parsing Shapes")
    shape_parser = ShapeParser(session, service_name)
    logger.debug("Parsing Client")
    client = parse_client(session, service_name, shape_parser)
    service_resource = parse_service_resource(session, service_name, shape_parser)

    result = ServicePackage(
        name=service_name.module_name,
        pypi_name=service_name.pypi_name,
        service_name=service_name,
        client=client,
        service_resource=service_resource,
    )

    for waiter_name in client.boto3_client.waiter_names:
        logger.debug(f"Parsing Waiter {waiter_name}")
        waiter = client.boto3_client.get_waiter(waiter_name)
        waiter_record = Waiter(
            name=f"{waiter.name}Waiter",
            docstring=(
                f"[Waiter.{waiter.name} documentation]"
                f"({service_name.doc_link}.Waiter.{waiter.name})"
            ),
            waiter_name=waiter_name,
            boto3_waiter=waiter,
        )

        wait_method = shape_parser.get_wait_method(waiter.name)
        wait_method.docstring = (
            f"[{waiter.name}.wait documentation]"
            f"({service_name.doc_link}.Waiter.{waiter.name}.wait)"
        )
        waiter_record.methods.append(wait_method)
        result.waiters.append(waiter_record)

    for paginator_name in shape_parser.get_paginator_names():
        logger.debug(f"Parsing Paginator {paginator_name}")
        operation_name = xform_name(paginator_name)
        paginator = client.boto3_client.get_paginator(operation_name)
        paginator_record = Paginator(
            name=f"{paginator_name}Paginator",
            operation_name=operation_name,
            boto3_paginator=paginator,
            docstring=(
                f"[Paginator.{paginator_name} documentation]"
                f"({service_name.doc_link}.Paginator.{paginator_name})"
            ),
        )

        paginate_method = shape_parser.get_paginate_method(paginator_name)
        paginate_method.docstring = (
            f"[{paginator_name}.paginate documentation]"
            f"({service_name.doc_link}.Paginator.{paginator_name}.paginate)"
        )
        paginator_record.methods.append(paginate_method)
        result.paginators.append(paginator_record)

    for paginator in result.paginators:
        method = paginator.get_client_method()
        result.client.methods.append(method)

    for waiter in result.waiters:
        method = waiter.get_client_method()
        result.client.methods.append(method)

    result.typed_dicts = result.extract_typed_dicts(result.get_types(), {})

    return result



"""
String to type annotation map that find type annotation by method and argument name.
"""
from typing import Dict, Optional

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_maps.typed_dicts import s3_copy_source_type, ec2_tag_type


OPERATION_TYPE_MAP: Dict[str, Dict[str, FakeAnnotation]] = {
    "UploadPartCopyRequest": {
        "CopySource": TypeSubscript(Type.Union, [Type.str, s3_copy_source_type])
    },
    "CopyObjectRequest": {
        "CopySource": TypeSubscript(Type.Union, [Type.str, s3_copy_source_type])
    },
    "CreateTagsRequest": {"Tags": TypeSubscript(Type.List, [ec2_tag_type])},
    "DeleteTagsRequest": {"Tags": TypeSubscript(Type.List, [ec2_tag_type])},
}


def get_argument_type_stub(
    operation_name: str, argument_name: str
) -> Optional[FakeAnnotation]:
    return OPERATION_TYPE_MAP.get(operation_name, {}).get(argument_name)



"""
Fake parser that produces `structures.ServiceModule` for master module and stubs.
"""
from boto3.session import Session
from botocore import xform_name

from mypy_boto3_builder.structures.client import Client
from mypy_boto3_builder.structures.service_resource import ServiceResource
from mypy_boto3_builder.structures.waiter import Waiter
from mypy_boto3_builder.structures.paginator import Paginator
from mypy_boto3_builder.structures.service_package import ServicePackage
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.parsers.boto3_utils import get_boto3_resource, get_boto3_client
from mypy_boto3_builder.utils.strings import get_class_prefix
from mypy_boto3_builder.parsers.shape_parser import ShapeParser


def parse_fake_service_package(
    session: Session, service_name: ServiceName
) -> ServicePackage:
    """
    Create fake boto3 service module structure.

    Used by stubs and master package.

    Arguments:
        session -- boto3 session.
        service_name -- Target service name.

    Returns:
        ServiceModule structure.
    """
    shape_parser = ShapeParser(session, service_name)

    result = ServicePackage(
        name=service_name.module_name,
        pypi_name=service_name.pypi_name,
        service_name=service_name,
        client=Client(),
    )

    boto3_client = get_boto3_client(session, service_name)
    boto3_resource = get_boto3_resource(session, service_name)

    if boto3_resource is not None:
        result.service_resource = ServiceResource()

    for waiter_name in boto3_client.waiter_names:
        real_class_name = get_class_prefix(waiter_name)
        waiter_class_name = f"{real_class_name}Waiter"
        result.waiters.append(Waiter(waiter_class_name, waiter_name=waiter_name))

    for paginator_name in shape_parser.get_paginator_names():
        operation_name = xform_name(paginator_name)
        result.paginators.append(
            Paginator(f"{paginator_name}Paginator", operation_name=operation_name)
        )

    return result



"""
String to type annotation map that find type annotation by argument name and type.
"""
from datetime import datetime
from typing import Dict

from mypy_boto3_builder.service_name import ServiceNameCatalog
from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_class import TypeClass


TYPE_MAP: Dict[str, FakeAnnotation] = {
    "bytes": Type.bytes,
    "blob": Type.bytes,
    "boolean": Type.bool,
    "function": TypeSubscript(Type.Callable, [Type.Ellipsis, Type.Any]),
    "botocore or boto3 Client": ExternalImport(
        source=ImportString("botocore", "client"), name="BaseClient"
    ),
    "datetime": TypeClass(datetime),
    "timestamp": TypeClass(datetime),
    "dict": TypeSubscript(Type.Dict, [Type.str, Type.Any]),
    "structure": TypeSubscript(Type.Dict, [Type.str, Type.Any]),
    "map": TypeSubscript(Type.Dict, [Type.str, Type.Any]),
    "float": Type.float,
    "double": Type.float,
    "int": Type.int,
    "integer": Type.int,
    "long": Type.int,
    "a file-like object": TypeSubscript(Type.IO, [Type.Any]),
    "seekable file-like object": TypeSubscript(Type.IO, [Type.Any]),
    "list": TypeSubscript(Type.List, [Type.Any]),
    "L{botocore.paginate.Paginator}": ExternalImport(
        source=ImportString("botocore", "paginate"), name="Paginator"
    ),
    ":py:class:`ResourceCollection`": ExternalImport(
        source=ImportString("boto3", "resources", "collection"),
        name="ResourceCollection",
    ),
    "JSON serializable": Type.str,
    "string": Type.str,
    "str": Type.str,
    "boto3.s3.transfer.TransferConfig": ExternalImport(
        source=ImportString("boto3", "s3", "transfer"), name="TransferConfig"
    ),
    "botocore.waiter.Waiter": ExternalImport(
        source=ImportString("botocore", "waiter"), name="Waiter"
    ),
    "bytes or seekable file-like object": TypeSubscript(
        Type.Union, [Type.bytes, Type.IO]
    ),
    "str or dict": TypeSubscript(Type.Union, [Type.str, Type.Dict]),
    "list(string)": TypeSubscript(Type.List, [Type.str]),
    "list of str": TypeSubscript(Type.List, [Type.str]),
    "None": Type.none,
    ":py:class:`ec2.NetworkAcl`": InternalImport("NetworkAcl", ServiceNameCatalog.ec2),
    ":py:class:`EC2.NetworkAcl`": InternalImport("NetworkAcl", ServiceNameCatalog.ec2),
    "list(:py:class:`ec2.InternetGateway`)": TypeSubscript(
        Type.List, [InternalImport("InternetGateway", ServiceNameCatalog.ec2)]
    ),
    ":py:class:`iam.UserPolicy`": InternalImport("UserPolicy", ServiceNameCatalog.iam),
    ":py:class:`IAM.UserPolicy`": InternalImport("UserPolicy", ServiceNameCatalog.iam),
    "list(:py:class:`iam.VirtualMfaDevice`)": TypeSubscript(
        Type.List, [InternalImport("VirtualMfaDevice", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`ec2.Image`)": TypeSubscript(
        Type.List, [InternalImport("Image", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`cloudwatch.Alarm`)": TypeSubscript(
        Type.List, [InternalImport("Alarm", ServiceNameCatalog.cloudwatch)]
    ),
    "list(:py:class:`opsworks.Layer`)": TypeSubscript(
        Type.List, [InternalImport("Layer", ServiceNameCatalog.opsworks)]
    ),
    ":py:class:`iam.GroupPolicy`": InternalImport("GroupPolicy", ServiceNameCatalog.iam),
    ":py:class:`IAM.GroupPolicy`": InternalImport("GroupPolicy", ServiceNameCatalog.iam),
    "list(:py:class:`iam.SigningCertificate`)": TypeSubscript(
        Type.List, [InternalImport("SigningCertificate", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`ec2.Volume`)": TypeSubscript(
        Type.List, [InternalImport("Volume", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`ec2.VpcPeeringConnection`)": TypeSubscript(
        Type.List, [InternalImport("VpcPeeringConnection", ServiceNameCatalog.ec2)]
    ),
    ":py:class:`ec2.Subnet`": InternalImport("Subnet", ServiceNameCatalog.ec2),
    ":py:class:`EC2.Subnet`": InternalImport("Subnet", ServiceNameCatalog.ec2),
    "list(:py:class:`iam.ServerCertificate`)": TypeSubscript(
        Type.List, [InternalImport("ServerCertificate", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`ec2.VpcAddress`)": TypeSubscript(
        Type.List, [InternalImport("VpcAddress", ServiceNameCatalog.ec2)]
    ),
    ":py:class:`sns.PlatformEndpoint`": InternalImport(
        "PlatformEndpoint", ServiceNameCatalog.sns
    ),
    ":py:class:`SNS.PlatformEndpoint`": InternalImport(
        "PlatformEndpoint", ServiceNameCatalog.sns
    ),
    ":py:class:`s3.MultipartUpload`": InternalImport(
        "MultipartUpload", ServiceNameCatalog.s3
    ),
    ":py:class:`glacier.MultipartUpload`": InternalImport(
        "MultipartUpload", ServiceNameCatalog.glacier
    ),
    ":py:class:`Glacier.MultipartUpload`": InternalImport(
        "MultipartUpload", ServiceNameCatalog.glacier
    ),
    ":py:class:`S3.MultipartUpload`": InternalImport(
        "MultipartUpload", ServiceNameCatalog.s3
    ),
    "list(:py:class:`ec2.Subnet`)": TypeSubscript(
        Type.List, [InternalImport("Subnet", ServiceNameCatalog.ec2)]
    ),
    ":py:class:`opsworks.Layer`": InternalImport("Layer", ServiceNameCatalog.opsworks),
    ":py:class:`OpsWorks.Layer`": InternalImport("Layer", ServiceNameCatalog.opsworks),
    "list(:py:class:`iam.MfaDevice`)": TypeSubscript(
        Type.List, [InternalImport("MfaDevice", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`glacier.Job`)": TypeSubscript(
        Type.List, [InternalImport("Job", ServiceNameCatalog.glacier)]
    ),
    "list(:py:class:`iam.RolePolicy`)": TypeSubscript(
        Type.List, [InternalImport("RolePolicy", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`iam.InstanceProfile`)": TypeSubscript(
        Type.List, [InternalImport("InstanceProfile", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`ec2.Instance`)": TypeSubscript(
        Type.List, [InternalImport("Instance", ServiceNameCatalog.ec2)]
    ),
    ":py:class:`glacier.Vault`": InternalImport("Vault", ServiceNameCatalog.glacier),
    ":py:class:`Glacier.Vault`": InternalImport("Vault", ServiceNameCatalog.glacier),
    ":py:class:`ec2.SecurityGroup`": InternalImport(
        "SecurityGroup", ServiceNameCatalog.ec2
    ),
    ":py:class:`EC2.SecurityGroup`": InternalImport(
        "SecurityGroup", ServiceNameCatalog.ec2
    ),
    ":py:class:`ec2.RouteTable`": InternalImport("RouteTable", ServiceNameCatalog.ec2),
    ":py:class:`EC2.RouteTable`": InternalImport("RouteTable", ServiceNameCatalog.ec2),
    "list(:py:class:`dynamodb.Table`)": TypeSubscript(
        Type.List, [InternalImport("Table", ServiceNameCatalog.dynamodb)]
    ),
    "list(:py:class:`ec2.Snapshot`)": TypeSubscript(
        Type.List, [InternalImport("Snapshot", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`sqs.Message`)": TypeSubscript(
        Type.List, [InternalImport("Message", ServiceNameCatalog.sqs)]
    ),
    "list(:py:class:`iam.Role`)": TypeSubscript(
        Type.List, [InternalImport("Role", ServiceNameCatalog.iam)]
    ),
    ":py:class:`glacier.Job`": InternalImport("Job", ServiceNameCatalog.glacier),
    ":py:class:`Glacier.Job`": InternalImport("Job", ServiceNameCatalog.glacier),
    "list(:py:class:`cloudwatch.Metric`)": TypeSubscript(
        Type.List, [InternalImport("Metric", ServiceNameCatalog.cloudwatch)]
    ),
    "list(:py:class:`iam.Policy`)": TypeSubscript(
        Type.List, [InternalImport("Policy", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`ec2.ClassicAddress`)": TypeSubscript(
        Type.List, [InternalImport("ClassicAddress", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`iam.User`)": TypeSubscript(
        Type.List, [InternalImport("User", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`iam.GroupPolicy`)": TypeSubscript(
        Type.List, [InternalImport("GroupPolicy", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`iam.PolicyVersion`)": TypeSubscript(
        Type.List, [InternalImport("PolicyVersion", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`sns.Topic`)": TypeSubscript(
        Type.List, [InternalImport("Topic", ServiceNameCatalog.sns)]
    ),
    ":py:class:`iam.LoginProfile`": InternalImport(
        "LoginProfile", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.LoginProfile`": InternalImport(
        "LoginProfile", ServiceNameCatalog.iam
    ),
    "list(:py:class:`iam.UserPolicy`)": TypeSubscript(
        Type.List, [InternalImport("UserPolicy", ServiceNameCatalog.iam)]
    ),
    "list(:py:class:`cloudformation.Event`)": TypeSubscript(
        Type.List, [InternalImport("Event", ServiceNameCatalog.cloudformation)]
    ),
    ":py:class:`cloudformation.Event`": InternalImport(
        "Event", ServiceNameCatalog.cloudformation
    ),
    ":py:class:`CloudFormation.Event`": InternalImport(
        "Event", ServiceNameCatalog.cloudformation
    ),
    "list(:py:class:`s3.MultipartUpload`)": TypeSubscript(
        Type.List, [InternalImport("MultipartUpload", ServiceNameCatalog.s3)]
    ),
    "list(:py:class:`glacier.MultipartUpload`)": TypeSubscript(
        Type.List, [InternalImport("MultipartUpload", ServiceNameCatalog.glacier)]
    ),
    "list(:py:class:`sns.Subscription`)": TypeSubscript(
        Type.List, [InternalImport("Subscription", ServiceNameCatalog.sns)]
    ),
    ":py:class:`iam.PolicyVersion`": InternalImport(
        "PolicyVersion", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.PolicyVersion`": InternalImport(
        "PolicyVersion", ServiceNameCatalog.iam
    ),
    "list(:py:class:`~boto3.resources.base.ServiceResource`)": TypeSubscript(
        Type.List,
        [
            ExternalImport(
                ImportString("boto3", "resources", "base"),
                name="ServiceResource",
                alias="Boto3ServiceResource",
            )
        ],
    ),
    "list(:py:class:`ec2.NetworkInterface`)": TypeSubscript(
        Type.List, [InternalImport("NetworkInterface", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`s3.ObjectVersion`)": TypeSubscript(
        Type.List, [InternalImport("ObjectVersion", ServiceNameCatalog.s3)]
    ),
    "list(:py:class:`ec2.SecurityGroup`)": TypeSubscript(
        Type.List, [InternalImport("SecurityGroup", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`sqs.Queue`)": TypeSubscript(
        Type.List, [InternalImport("Queue", ServiceNameCatalog.sqs)]
    ),
    "list(:py:class:`ec2.PlacementGroup`)": TypeSubscript(
        Type.List, [InternalImport("PlacementGroup", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`ec2.Vpc`)": TypeSubscript(
        Type.List, [InternalImport("Vpc", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`ec2.RouteTable`)": TypeSubscript(
        Type.List, [InternalImport("RouteTable", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`glacier.Vault`)": TypeSubscript(
        Type.List, [InternalImport("Vault", ServiceNameCatalog.glacier)]
    ),
    "list(:py:class:`iam.Group`)": TypeSubscript(
        Type.List, [InternalImport("Group", ServiceNameCatalog.iam)]
    ),
    ":py:class:`iam.Group`": TypeSubscript(
        Type.List, [InternalImport("Group", ServiceNameCatalog.iam)]
    ),
    ":py:class:`ec2.Image`": InternalImport("Image", ServiceNameCatalog.ec2),
    ":py:class:`EC2.Image`": InternalImport("Image", ServiceNameCatalog.ec2),
    ":py:class:`ec2.Route`": InternalImport("Route", ServiceNameCatalog.ec2),
    ":py:class:`EC2.Route`": InternalImport("Route", ServiceNameCatalog.ec2),
    ":py:class:`ec2.VpcPeeringConnection`": InternalImport(
        "VpcPeeringConnection", ServiceNameCatalog.ec2
    ),
    ":py:class:`EC2.VpcPeeringConnection`": InternalImport(
        "VpcPeeringConnection", ServiceNameCatalog.ec2
    ),
    "list(:py:class:`cloudformation.Stack`)": TypeSubscript(
        Type.List, [InternalImport("Stack", ServiceNameCatalog.cloudformation)]
    ),
    "list(:py:class:`opsworks.Stack`)": TypeSubscript(
        Type.List, [InternalImport("Stack", ServiceNameCatalog.opsworks)]
    ),
    ":py:class:`iam.MfaDevice`": InternalImport("MfaDevice", ServiceNameCatalog.iam),
    ":py:class:`IAM.MfaDevice`": InternalImport("MfaDevice", ServiceNameCatalog.iam),
    "list(:py:class:`s3.Bucket`)": TypeSubscript(
        Type.List, [InternalImport("Bucket", ServiceNameCatalog.s3)]
    ),
    "list(:py:class:`sns.PlatformEndpoint`)": TypeSubscript(
        Type.List, [InternalImport("PlatformEndpoint", ServiceNameCatalog.sns)]
    ),
    ":py:class:`ec2.Snapshot`": InternalImport("Snapshot", ServiceNameCatalog.ec2),
    ":py:class:`EC2.Snapshot`": InternalImport("Snapshot", ServiceNameCatalog.ec2),
    "list(:py:class:`ec2.DhcpOptions`)": TypeSubscript(
        Type.List, [InternalImport("DhcpOptions", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`ec2.NetworkAcl`)": TypeSubscript(
        Type.List, [InternalImport("NetworkAcl", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`ec2.KeyPairInfo`)": TypeSubscript(
        Type.List, [InternalImport("KeyPairInfo", ServiceNameCatalog.ec2)]
    ),
    "list(:py:class:`cloudformation.StackResourceSummary`)": TypeSubscript(
        Type.List,
        [InternalImport("StackResourceSummary", ServiceNameCatalog.cloudformation)],
    ),
    ":py:class:`dynamodb.Table`": InternalImport("Table", ServiceNameCatalog.dynamodb),
    ":py:class:`DynamoDB.Table`": InternalImport("Table", ServiceNameCatalog.dynamodb),
    ":py:class:`iam.AccessKeyPair`": InternalImport(
        "AccessKeyPair", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.AccessKeyPair`": InternalImport(
        "AccessKeyPair", ServiceNameCatalog.iam
    ),
    "list(:py:class:`iam.SamlProvider`)": TypeSubscript(
        Type.List, [InternalImport("SamlProvider", ServiceNameCatalog.iam)]
    ),
    ":py:class:`glacier.Archive`": InternalImport("Archive", ServiceNameCatalog.glacier),
    ":py:class:`Glacier.Archive`": InternalImport("Archive", ServiceNameCatalog.glacier),
    ":py:class:`ec2.NetworkInterface`": InternalImport(
        "NetworkInterface", ServiceNameCatalog.ec2
    ),
    ":py:class:`EC2.NetworkInterface`": InternalImport(
        "NetworkInterface", ServiceNameCatalog.ec2
    ),
    "list(:py:class:`iam.AccessKey`)": TypeSubscript(
        Type.List, [InternalImport("AccessKey", ServiceNameCatalog.iam)]
    ),
    ":py:class:`sns.Subscription`": InternalImport(
        "Subscription", ServiceNameCatalog.sns
    ),
    ":py:class:`SNS.Subscription`": InternalImport(
        "Subscription", ServiceNameCatalog.sns
    ),
    "list(:py:class:`s3.MultipartUploadPart`)": TypeSubscript(
        Type.List, [InternalImport("MultipartUploadPart", ServiceNameCatalog.s3)]
    ),
    ":py:class:`iam.ServerCertificate`": InternalImport(
        "ServerCertificate", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.ServerCertificate`": InternalImport(
        "ServerCertificate", ServiceNameCatalog.iam
    ),
    "list(:py:class:`ec2.Tag`)": TypeSubscript(
        Type.List, [InternalImport("Tag", ServiceNameCatalog.ec2)]
    ),
    ":py:class:`cloudwatch.Alarm`": InternalImport(
        "Alarm", ServiceNameCatalog.cloudwatch
    ),
    ":py:class:`CloudWatch.Alarm`": InternalImport(
        "Alarm", ServiceNameCatalog.cloudwatch
    ),
    ":py:class:`EC2.PlacementGroup`": InternalImport(
        "PlacementGroup", ServiceNameCatalog.ec2
    ),
    ":py:class:`ec2.PlacementGroup`": InternalImport(
        "PlacementGroup", ServiceNameCatalog.ec2
    ),
    ":py:class:`EC2.Vpc`": InternalImport("Vpc", ServiceNameCatalog.ec2),
    ":py:class:`ec2.Vpc`": InternalImport("Vpc", ServiceNameCatalog.ec2),
    ":py:class:`S3.BucketVersioning`": InternalImport(
        "BucketVersioning", ServiceNameCatalog.s3
    ),
    ":py:class:`IAM.User`": InternalImport("User", ServiceNameCatalog.iam),
    ":py:class:`iam.User`": InternalImport("User", ServiceNameCatalog.iam),
    ":py:class:`sns.Topic`": InternalImport("Topic", ServiceNameCatalog.sns),
    ":py:class:`SNS.Topic`": InternalImport("Topic", ServiceNameCatalog.sns),
    ":py:class:`iam.Policy`": InternalImport("Policy", ServiceNameCatalog.iam),
    ":py:class:`IAM.Policy`": InternalImport("Policy", ServiceNameCatalog.iam),
    ":py:class:`S3.BucketCors`": InternalImport("BucketCors", ServiceNameCatalog.s3),
    ":py:class:`OpsWorks.Stack`": InternalImport("Stack", ServiceNameCatalog.opsworks),
    ":py:class:`CloudFormation.Stack`": InternalImport(
        "Stack", ServiceNameCatalog.cloudformation
    ),
    ":py:class:`opsworks.Stack`": InternalImport("Stack", ServiceNameCatalog.opsworks),
    ":py:class:`cloudformation.Stack`": InternalImport(
        "Stack", ServiceNameCatalog.cloudformation
    ),
    ":py:class:`IAM.SigningCertificate`": InternalImport(
        "SigningCertificate", ServiceNameCatalog.iam
    ),
    ":py:class:`iam.SigningCertificate`": InternalImport(
        "SigningCertificate", ServiceNameCatalog.iam
    ),
    ":py:class:`S3.ObjectVersion`": InternalImport(
        "ObjectVersion", ServiceNameCatalog.s3
    ),
    ":py:class:`S3.BucketPolicy`": InternalImport("BucketPolicy", ServiceNameCatalog.s3),
    ":py:class:`EC2.RouteTableAssociation`": InternalImport(
        "RouteTableAssociation", ServiceNameCatalog.ec2
    ),
    ":py:class:`ec2.RouteTableAssociation`": InternalImport(
        "RouteTableAssociation", ServiceNameCatalog.ec2
    ),
    ":py:class:`IAM.RolePolicy`": InternalImport("RolePolicy", ServiceNameCatalog.iam),
    ":py:class:`IAM.CurrentUser`": InternalImport("CurrentUser", ServiceNameCatalog.iam),
    ":py:class:`EC2.InternetGateway`": InternalImport(
        "InternetGateway", ServiceNameCatalog.ec2
    ),
    ":py:class:`ec2.InternetGateway`": InternalImport(
        "InternetGateway", ServiceNameCatalog.ec2
    ),
    ":py:class:`sns.PlatformApplication`": InternalImport(
        "PlatformApplication", ServiceNameCatalog.sns
    ),
    ":py:class:`SNS.PlatformApplication`": InternalImport(
        "PlatformApplication", ServiceNameCatalog.sns
    ),
    ":py:class:`CloudWatch.Metric`": InternalImport(
        "Metric", ServiceNameCatalog.cloudwatch
    ),
    ":py:class:`IAM.Group`": InternalImport("Group", ServiceNameCatalog.iam),
    ":py:class:`OpsWorks.StackSummary`": InternalImport(
        "StackSummary", ServiceNameCatalog.opsworks
    ),
    ":py:class:`IAM.AssumeRolePolicy`": InternalImport(
        "AssumeRolePolicy", ServiceNameCatalog.iam
    ),
    ":py:class:`S3.Object`": InternalImport("Object", ServiceNameCatalog.s3),
    ":py:class:`s3.Object`": InternalImport("Object", ServiceNameCatalog.s3),
    ":py:class:`CloudFormation.StackResourceSummary`": InternalImport(
        "StackResourceSummary", ServiceNameCatalog.cloudformation
    ),
    ":py:class:`S3.BucketWebsite`": InternalImport(
        "BucketWebsite", ServiceNameCatalog.s3
    ),
    ":py:class:`SQS.Queue`": InternalImport("Queue", ServiceNameCatalog.sqs),
    ":py:class:`sqs.Queue`": InternalImport("Queue", ServiceNameCatalog.sqs),
    ":py:class:`S3.MultipartUploadPart`": InternalImport(
        "MultipartUploadPart", ServiceNameCatalog.s3
    ),
    ":py:class:`ec2.KeyPairInfo`": InternalImport("KeyPairInfo", ServiceNameCatalog.ec2),
    ":py:class:`EC2.KeyPairInfo`": InternalImport("KeyPairInfo", ServiceNameCatalog.ec2),
    ":py:class:`iam.InstanceProfile`": InternalImport(
        "InstanceProfile", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.InstanceProfile`": InternalImport(
        "InstanceProfile", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.AccessKey`": InternalImport("AccessKey", ServiceNameCatalog.iam),
    ":py:class:`IAM.SamlProvider`": InternalImport(
        "SamlProvider", ServiceNameCatalog.iam
    ),
    ":py:class:`iam.SamlProvider`": InternalImport(
        "SamlProvider", ServiceNameCatalog.iam
    ),
    ":py:class:`EC2.Tag`": InternalImport("Tag", ServiceNameCatalog.ec2),
    ":py:class:`S3.BucketLifecycleConfiguration`": InternalImport(
        "BucketLifecycleConfiguration", ServiceNameCatalog.s3
    ),
    ":py:class:`iam.AccountPasswordPolicy`": InternalImport(
        "AccountPasswordPolicy", ServiceNameCatalog.iam
    ),
    ":py:class:`IAM.AccountPasswordPolicy`": InternalImport(
        "AccountPasswordPolicy", ServiceNameCatalog.iam
    ),
    ":py:class:`S3.BucketNotification`": InternalImport(
        "BucketNotification", ServiceNameCatalog.s3
    ),
    ":py:class:`CloudFormation.StackResource`": InternalImport(
        "StackResource", ServiceNameCatalog.cloudformation
    ),
    ":py:class:`EC2.Instance`": InternalImport("Instance", ServiceNameCatalog.ec2),
    ":py:class:`S3.BucketTagging`": InternalImport(
        "BucketTagging", ServiceNameCatalog.s3
    ),
    ":py:class:`IAM.AccountSummary`": InternalImport(
        "AccountSummary", ServiceNameCatalog.iam
    ),
    ":py:class:`EC2.Volume`": InternalImport("Volume", ServiceNameCatalog.ec2),
    ":py:class:`ec2.Volume`": InternalImport("Volume", ServiceNameCatalog.ec2),
    ":py:class:`SQS.Message`": InternalImport("Message", ServiceNameCatalog.sqs),
    ":py:class:`ec2.KeyPair`": InternalImport("KeyPair", ServiceNameCatalog.ec2),
    ":py:class:`S3.BucketAcl`": InternalImport("BucketAcl", ServiceNameCatalog.s3),
    ":py:class:`S3.BucketRequestPayment`": InternalImport(
        "BucketRequestPayment", ServiceNameCatalog.s3
    ),
    ":py:class:`IAM.Role`": InternalImport("Role", ServiceNameCatalog.iam),
    ":py:class:`iam.Role`": InternalImport("Role", ServiceNameCatalog.iam),
    ":py:class:`IAM.VirtualMfaDevice`": InternalImport(
        "VirtualMfaDevice", ServiceNameCatalog.iam
    ),
    ":py:class:`iam.VirtualMfaDevice`": InternalImport(
        "VirtualMfaDevice", ServiceNameCatalog.iam
    ),
    ":py:class:`Glacier.Account`": InternalImport("Account", ServiceNameCatalog.glacier),
    ":py:class:`S3.BucketLifecycle`": InternalImport(
        "BucketLifecycle", ServiceNameCatalog.s3
    ),
    ":py:class:`EC2.DhcpOptions`": InternalImport("DhcpOptions", ServiceNameCatalog.ec2),
    ":py:class:`ec2.DhcpOptions`": InternalImport("DhcpOptions", ServiceNameCatalog.ec2),
    ":py:class:`S3.BucketLogging`": InternalImport(
        "BucketLogging", ServiceNameCatalog.s3
    ),
    ":py:class:`Glacier.Notification`": InternalImport(
        "Notification", ServiceNameCatalog.glacier
    ),
    ":py:class:`EC2.ClassicAddress`": InternalImport(
        "ClassicAddress", ServiceNameCatalog.ec2
    ),
    ":py:class:`s3.Bucket`": InternalImport("Bucket", ServiceNameCatalog.s3),
    ":py:class:`S3.Bucket`": InternalImport("Bucket", ServiceNameCatalog.s3),
    ":py:class:`EC2.VpcAddress`": InternalImport("VpcAddress", ServiceNameCatalog.ec2),
    ":py:class:`S3.ObjectSummary`": InternalImport(
        "ObjectSummary", ServiceNameCatalog.s3
    ),
    ":py:class:`EC2.NetworkInterfaceAssociation`": InternalImport(
        "NetworkInterfaceAssociation", ServiceNameCatalog.ec2
    ),
    ":py:class:`S3.ObjectAcl`": InternalImport("ObjectAcl", ServiceNameCatalog.s3),
    "list(:py:class:`sns.PlatformApplication`)": TypeSubscript(
        Type.List, [InternalImport("PlatformApplication", ServiceNameCatalog.sns)]
    ),
    "list(:py:class:`s3.ObjectSummary`)": TypeSubscript(
        Type.List, [InternalImport("ObjectSummary", ServiceNameCatalog.s3)]
    ),
    "botocore.waiter.SingleWaiterConfig": ExternalImport(
        source=ImportString("botocore", "waiter"), name="SingleWaiterConfig"
    ),
    "callable": TypeSubscript(Type.Callable, [Type.Ellipsis, Type.Any]),
}



"""
Parent class for all type annotation wrappers.
"""
from __future__ import annotations

from abc import abstractmethod
from typing import Set, Any
from functools import total_ordering

from mypy_boto3_builder.import_helpers.import_record import ImportRecord


@total_ordering
class FakeAnnotation:
    """
    Parent class for all type annotation wrappers.
    """

    def __hash__(self) -> int:
        return hash(self.render())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FakeAnnotation):
            raise ValueError(f"Cannot compare FakeAnnotation with {other}")

        return str(self) == str(other)

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, FakeAnnotation):
            raise ValueError(f"Cannot compare FakeAnnotation with {other}")

        return not self == other

    def __gt__(self, other: FakeAnnotation) -> bool:
        return str(self) > str(other)

    def __str__(self) -> str:
        return self.render()

    @abstractmethod
    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.
        """

    @abstractmethod
    def get_import_record(self) -> ImportRecord:
        """
        Get import record required for using type annotation.
        """

    def get_types(self) -> Set[FakeAnnotation]:
        """
        Get all used type annotations recursively including self.
        """
        return {self}

    def remove_children(self) -> None:
        """
        Remove all children from `TypeSubscript` or `TypeTypedDict` annotation.
        """

    def add_child(self, child: FakeAnnotation) -> None:
        """
        Add new child to `TypeSubscript` or `TypeTypedDict` annotation.
        """

    def add_literal_child(self, child: Any) -> None:
        """
        Add new child to `TypeLiteral` annotation.
        """

    def is_dict(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether type annotation is `Dict` or `TypedDict`.
        """
        return False

    def is_typed_dict(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether type annotation is `TypedDict`.
        """
        return False

    def is_list(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether type annotation is `List`.
        """
        return False

    def is_literal(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether type annotation is `Literal`.
        """
        return False

    def is_potential_literal(self) -> bool:  # pylint: disable=no-self-use
        """
        Whether type annotation can be a `Literal`.
        """
        return False

    @abstractmethod
    def copy(self) -> FakeAnnotation:
        """
        Create a copy of type annotation wrapper.
        """



"""
Boto3 ServiceResource collections parser, produces `structures.Collection`.
"""
from typing import List

from boto3.resources.base import ServiceResource as Boto3ServiceResource

from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.utils.strings import get_class_prefix
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript


def parse_collections(
    parent_name: str, resource: Boto3ServiceResource, service_name: ServiceName
) -> List[Collection]:
    """
    Extract collections from boto3 resource.

    Arguments:
        resource -- boto3 service resource.

    Returns:
        A list of Collection structures.
    """
    result: List[Collection] = []
    for collection in resource.meta.resource_model.collections:
        if not collection.resource:
            continue
        object_class_name = collection.resource.type
        collection_record = Collection(
            name=f"{parent_name}{get_class_prefix(collection.name)}Collection",
            parent_name=parent_name,
            attribute_name=collection.name,
            docstring=(
                f"[{parent_name}.{collection.name} documentation]"
                f"({service_name.doc_link}.{parent_name}.{collection.name})"
            ),
            type=InternalImport(name=collection.name, service_name=service_name),
        )
        self_type = InternalImport(
            name=collection_record.name, service_name=service_name
        )

        collection_record.methods.append(
            Method("all", [Argument("cls", None)], self_type)
        )
        collection_record.methods.append(
            Method(
                "filter",
                [
                    Argument("cls", None),
                    Argument("Delimiter", Type.str, Type.none),
                    Argument("EncodingType", Type.str, Type.none),
                    Argument("KeyMarker", Type.str, Type.none),
                    Argument("MaxUploads", Type.int, Type.none),
                    Argument("Prefix", Type.str, Type.none),
                    Argument("UploadIdMarker", Type.str, Type.none),
                ],
                self_type,
                decorators=[Type.classmethod],
            )
        )
        collection_record.methods.append(
            Method(
                "limit",
                [Argument("cls", None), Argument("count", Type.int)],
                self_type,
                decorators=[Type.classmethod],
            )
        )
        collection_record.methods.append(
            Method(
                "page_size",
                [Argument("cls", None), Argument("count", Type.int)],
                self_type,
                decorators=[Type.classmethod],
            )
        )
        collection_record.methods.append(
            Method(
                "pages",
                [Argument("cls", None)],
                TypeSubscript(
                    Type.List,
                    [InternalImport(name=object_class_name, service_name=service_name)],
                ),
                decorators=[Type.classmethod],
            )
        )

        result.append(collection_record)
    return result



from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_maps.typed_dicts import ec2_tag_type


__all__ = ("DOCLESS_METHOD_ARGUMENT_MAP",)


DOCLESS_METHOD_ARGUMENT_MAP = {
    "ServiceResource.create_tags": [
        Argument("self", None),
        Argument("Resources", TypeSubscript(Type.List, [Type.Any]),),
        Argument("Tags", TypeSubscript(Type.List, [ec2_tag_type]),),
        Argument("DryRun", Type.bool, TypeConstant(False)),
    ]
}



"""
Provides compatibility between `typing` and `typing_extensions`.
"""
import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict, Literal


__all__ = (
    "TypedDict",
    "Literal",
)



"""
Method or function argument.
"""
from dataclasses import dataclass
from typing import Set, Optional

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant


@dataclass
class Argument:
    """
    Method or function argument.

    Attributes:
        name -- Argument name.
        type -- Argument type annotation.
        value -- Default argument value.
        prefix -- Used for starargs.
    """

    name: str
    type: Optional[FakeAnnotation]
    default: Optional[TypeConstant] = None
    prefix: str = ""

    def get_types(self) -> Set[FakeAnnotation]:
        types: Set[FakeAnnotation] = set()
        if self.type is not None:
            types.update(self.type.get_types())
        if self.default is not None:
            types.update(self.default.get_types())

        return types

    @property
    def required(self) -> bool:
        return self.default is None



"""
Boto3 ServiceResource or Resource collection.
"""
from dataclasses import dataclass, field
from typing import Set, List

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.structures.class_record import ClassRecord


@dataclass
class Collection(ClassRecord):
    """
    Boto3 ServiceResource or Resource collection.
    """

    attribute_name: str = "collection"
    parent_name: str = "Parent"
    type: FakeAnnotation = Type.Any
    bases: List[FakeAnnotation] = field(
        default_factory=lambda: [
            ExternalImport(
                source=ImportString("boto3", "resources", "collection"),
                name="ResourceCollection",
            )
        ]
    )

    def get_types(self) -> Set[FakeAnnotation]:
        types = super().get_types()
        types.update(self.type.get_types())
        return types



# pylint: disable=unused-argument,multiple-statements,no-self-use
from __future__ import annotations

import sys
from decimal import Context
from typing import Tuple, Any, Dict

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

DynamoDBDataType = Literal["S", "N", "B", "SS", "NS", "BS", "NULL", "BOOL", "M", "L"]

STRING: Literal["S"]
NUMBER: Literal["N"]
BINARY: Literal["B"]
STRING_SET: Literal["SS"]
NUMBER_SET: Literal["NS"]
BINARY_SET: Literal["BS"]
NULL: Literal["NULL"]
BOOLEAN: Literal["BOOL"]
MAP: Literal["M"]
LIST: Literal["L"]

DYNAMODB_CONTEXT: Context

BINARY_TYPES: Tuple[type]

class Binary:
    def __init__(self, value: Any) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __ne__(self, other: Any) -> bool: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __hash__(self) -> int: ...

class TypeSerializer:
    def serialize(self, value: Any) -> Dict[DynamoDBDataType, Any]: ...

class TypeDeserializer:
    def deserialize(self, value: Dict[DynamoDBDataType, Any]) -> Any: ...



"""
Wrapper for simple type annotation like `str` or `Dict`.
"""
from __future__ import annotations

from typing import Union, Optional, Any, Dict, List, Callable, IO, overload

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation


class TypeAnnotation(FakeAnnotation):
    """
    Wrapper for simple type annotation like `str` or `Dict`.

    Arguments:
        wrapped_type -- Original type annotation.
    """

    supported_types = [
        Union,
        Any,
        Dict,
        List,
        Optional,
        Callable,
        IO,
        overload,
    ]

    def __init__(self, wrapped_type: Any) -> None:
        if isinstance(wrapped_type, FakeAnnotation):
            raise ValueError(f"Cannot wrap FakeAnnotation: {wrapped_type}")
        if wrapped_type not in self.supported_types:
            raise ValueError(f"Cannot wrap {wrapped_type}")

        self.wrapped_type = wrapped_type

    def render(self, parent_name: str = "") -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        return self.get_import_name()

    def get_import_name(self) -> str:
        if self.wrapped_type is IO:
            return "IO"
        if self.wrapped_type is Callable:
            return "Callable"
        if self.wrapped_type is overload:
            return "overload"
        return str(getattr(self.wrapped_type, "_name"))

    def get_import_record(self) -> ImportRecord:
        return ImportRecord(source=ImportString("typing"), name=self.get_import_name())

    def is_dict(self) -> bool:
        return self.wrapped_type is Dict

    def is_list(self) -> bool:
        return self.wrapped_type is List

    def copy(self) -> TypeAnnotation:
        """
        Create a copy of type annotation wrapper.
        """
        return TypeAnnotation(self.wrapped_type)



"""
boto3-stubs package writer.
"""
from pathlib import Path
import shutil
from typing import List
import filecmp

from boto3 import __version__ as boto3_version

from mypy_boto3_builder.structures.boto3_stubs_package import Boto3StubsPackage
from mypy_boto3_builder.version import __version__ as version
from mypy_boto3_builder.writers.utils import render_jinja2_template, blackify
from mypy_boto3_builder.constants import BOTO3_STUBS_STATIC_PATH


def write_boto3_stubs_package(
    package: Boto3StubsPackage, output_path: Path
) -> List[Path]:
    modified_paths: List[Path] = []
    package_path = output_path / package.name

    output_path.mkdir(exist_ok=True)
    package_path.mkdir(exist_ok=True)

    templates_path = Path("boto3-stubs")
    module_templates_path = templates_path / "boto3-stubs"
    file_paths = [
        (output_path / "setup.py", templates_path / "setup.py.jinja2"),
        (output_path / "README.md", templates_path / "README.md.jinja2"),
        (package_path / "py.typed", module_templates_path / "py.typed.jinja2"),
        (package_path / "__init__.pyi", module_templates_path / "__init__.pyi.jinja2"),
        (package_path / "session.pyi", module_templates_path / "session.pyi.jinja2"),
        (package_path / "__init__.py", module_templates_path / "__init__.py.jinja2"),
        (package_path / "version.py", module_templates_path / "version.py.jinja2"),
    ]

    for file_path, template_path in file_paths:
        content = render_jinja2_template(template_path, package=package)
        content = blackify(content, file_path)
        if not file_path.exists() or file_path.read_text() != content:
            modified_paths.append(file_path)
            file_path.write_text(content)

    for static_path in BOTO3_STUBS_STATIC_PATH.glob("**/*.pyi"):
        relative_output_path = static_path.relative_to(BOTO3_STUBS_STATIC_PATH)
        file_path = package_path / relative_output_path
        file_path.parent.mkdir(exist_ok=True)
        if file_path.exists() and filecmp.cmp(
            static_path.as_posix(), file_path.as_posix()
        ):
            continue

        shutil.copy(static_path, file_path)
        modified_paths.append(file_path)

    return modified_paths



"""
Boto3 ServiceResource sub-Resource.
"""
from dataclasses import dataclass, field
from typing import Set, List, Optional

from mypy_boto3_builder.import_helpers.import_string import ImportString
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.collection import Collection
from mypy_boto3_builder.structures.method import Method


@dataclass
class Resource(ClassRecord):
    """
    Boto3 ServiceResource sub-Resource.
    """

    bases: List[FakeAnnotation] = field(
        default_factory=lambda: [
            ExternalImport(
                source=ImportString("boto3", "resources", "base"),
                name="ServiceResource",
                alias="Boto3ServiceResource",
            )
        ]
    )
    collections: List[Collection] = field(default_factory=lambda: [])
    init_method: Optional[Method] = None

    def get_types(self) -> Set[FakeAnnotation]:
        types = super().get_types()
        for collection in self.collections:
            types.update(collection.get_types())
        return types



"""
Boto3 client Waiter.
"""
from dataclasses import dataclass, field
from typing import List

from botocore.waiter import Waiter as Boto3Waiter

from mypy_boto3_builder.import_helpers.internal_import_record import (
    InternalImportRecord,
)
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_class import TypeClass
from mypy_boto3_builder.type_annotations.type_literal import TypeLiteral
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.structures.class_record import ClassRecord
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.argument import Argument


@dataclass
class Waiter(ClassRecord):
    """
    Boto3 client Waiter.
    """

    waiter_name: str = "waiter_name"
    boto3_waiter: Boto3Waiter = None
    bases: List[FakeAnnotation] = field(
        default_factory=lambda: [TypeClass(Boto3Waiter, alias="Boto3Waiter")]
    )

    def get_import_record(self) -> InternalImportRecord:
        return InternalImportRecord(ServiceModuleName.waiter, name=self.name)

    def get_client_method(self) -> Method:
        return Method(
            name="get_waiter",
            decorators=[Type.overload],
            docstring=self.docstring,
            arguments=[
                Argument("self", None),
                Argument("waiter_name", TypeLiteral(self.waiter_name)),
            ],
            return_type=InternalImport(self.name, module_name=ServiceModuleName.waiter),
        )



# pylint: disable=unused-argument,multiple-statements,super-init-not-called,no-self-use
import logging
from typing import Dict, Type

from botocore.hooks import HierarchicalEmitter

from boto3.resources.base import ServiceResource
from boto3.utils import ServiceContext

logger: logging.Logger

class ResourceFactory:
    def __init__(self, emitter: HierarchicalEmitter) -> None: ...
    def load_from_definition(
        self,
        resource_name: str,
        single_resource_json_definition: Dict,
        service_context: ServiceContext,
    ) -> Type[ServiceResource]: ...



"""
Master package writer.
"""
from pathlib import Path
from typing import List

from mypy_boto3_builder.structures.master_package import MasterPackage
from mypy_boto3_builder.version import __version__ as version
from mypy_boto3_builder.writers.utils import render_jinja2_template, blackify


def write_master_package(package: MasterPackage, output_path: Path) -> List[Path]:
    modified_paths: List[Path] = []
    package_path = output_path / package.name

    output_path.mkdir(exist_ok=True)
    package_path.mkdir(exist_ok=True)

    templates_path = Path("master")
    module_templates_path = templates_path / "master"
    file_paths = [
        (output_path / "setup.py", templates_path / "setup.py.jinja2"),
        (output_path / "README.md", templates_path / "README.md.jinja2"),
        (package_path / "__init__.py", module_templates_path / "__init__.py.jinja2"),
        (package_path / "__main__.py", module_templates_path / "__main__.py.jinja2"),
        (package_path / "py.typed", module_templates_path / "py.typed.jinja2"),
        (package_path / "version.py", module_templates_path / "version.py.jinja2"),
        (package_path / "main.py", module_templates_path / "main.py.jinja2"),
    ]

    for file_path, template_path in file_paths:
        content = render_jinja2_template(template_path, package=package)
        content = blackify(content, file_path)

        if not file_path.exists() or file_path.read_text() != content:
            modified_paths.append(file_path)
            file_path.write_text(content)

    return modified_paths



"""
Constants and paths.
"""
from pathlib import Path

# Master module name
MODULE_NAME = "mypy_boto3"

# PyPI module name
PYPI_NAME = "mypy-boto3"

# Random region to initialize services
DUMMY_REGION = "us-west-2"

# Jinja2 templates for boto3-stubs
TEMPLATES_PATH = Path(__file__).parent / "templates"

# Static *.pyi files for boto3-stubs
BOTO3_STUBS_STATIC_PATH = Path(__file__).parent / "boto3_stubs_static"

# Boto3 stubs module name
BOTO3_STUBS_NAME = "boto3-stubs"

# Max line length for boto3 docs
LINE_LENGTH = 100

# type defs module name
TYPE_DEFS_NAME = "type_defs"



"""
String to type annotation map that find type annotation in botocore syntax
"""
from typing import Dict

from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type import Type


SYNTAX_TYPE_MAP: Dict[str, FakeAnnotation] = {
    "'string'": Type.str,
    "b'bytes'": Type.bytes,
    "123": Type.int,
    "123.0": Type.float,
    "123.4": Type.float,
    "None": Type.none,
    "True": Type.bool,
    "False": Type.bool,
    "file": Type.IO,
    "...": Type.Any,
}



"""
Converter of function argspec to `Argument` list.
"""
import inspect
from typing import List, Optional
from types import FunctionType


from mypy_boto3_builder.structures.argument import Argument
from mypy_boto3_builder.type_annotations.type_annotation import TypeAnnotation
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_constant import TypeConstant
from mypy_boto3_builder.type_maps.method_type_map import METHOD_TYPE_MAP


class ArgSpecParser:
    """
    Converter of function argspec to `Argument` list.
    """

    def __init__(self, prefix: str) -> None:
        self.prefix = prefix

    @staticmethod
    def _get_arguments_from_argspec(func: FunctionType) -> List[Argument]:
        arguments: List[Argument] = []
        argspec = inspect.getfullargspec(func)
        for argument_name in argspec.args:
            if argument_name == "factory_self":
                argument_name = "self"
            type_annotation: Optional[TypeAnnotation] = Type.Any
            if not arguments and argument_name in ("self", "cls"):
                type_annotation = None
            arguments.append(Argument(argument_name, type_annotation))
        if argspec.defaults:
            for index, default_value in enumerate(argspec.defaults):
                argument_index = len(arguments) - len(argspec.defaults) + index
                arguments[argument_index].default = TypeConstant(default_value)

        if argspec.varargs:
            arguments.append(Argument(argspec.varargs, Type.Any, prefix="*"))
        for argument_name in argspec.kwonlyargs:
            arguments.append(Argument(argument_name, Type.Any))
        if argspec.kwonlydefaults:
            for argument_name, default_value in argspec.kwonlydefaults:
                for argument in arguments:
                    if argument.name != argument_name:
                        continue
                    argument.default = TypeConstant(default_value)
                    break
        if argspec.varkw:
            arguments.append(Argument(argspec.varkw, Type.Any, prefix="**"))
        return arguments

    def get_function_arguments(self, func: FunctionType) -> List[Argument]:
        func_name = func.__name__
        arguments = self._get_arguments_from_argspec(func)

        for argument in arguments:
            if argument.type is not Type.Any:
                continue

            method_type = f"{func_name}: {argument.name}"
            if method_type in METHOD_TYPE_MAP:
                argument.type = METHOD_TYPE_MAP[method_type]

        return arguments



"""
Boto3 client parser, produces `structures.Client`.
"""
from boto3.session import Session
from botocore.errorfactory import ClientExceptionsFactory
from botocore.exceptions import ClientError

from mypy_boto3_builder.structures.client import Client
from mypy_boto3_builder.structures.attribute import Attribute
from mypy_boto3_builder.service_name import ServiceName
from mypy_boto3_builder.enums.service_module_name import ServiceModuleName
from mypy_boto3_builder.type_annotations.type_class import TypeClass
from mypy_boto3_builder.type_annotations.internal_import import InternalImport
from mypy_boto3_builder.parsers.helpers import (
    parse_method,
    get_public_methods,
)
from mypy_boto3_builder.parsers.boto3_utils import get_boto3_client
from mypy_boto3_builder.parsers.shape_parser import ShapeParser


def parse_client(
    session: Session, service_name: ServiceName, shape_parser: ShapeParser
) -> Client:
    """
    Parse boto3 client to a structure.

    Arguments:
        session -- boto3 session.
        service_name -- Target service name.

    Returns:
        Client structure.
    """
    client = get_boto3_client(session, service_name)
    public_methods = get_public_methods(client)

    # remove methods that will be overriden
    if "get_paginator" in public_methods:
        del public_methods["get_paginator"]
    if "get_waiter" in public_methods:
        del public_methods["get_waiter"]

    result = Client(
        service_name=service_name,
        boto3_client=client,
        docstring=(
            f"[{service_name.class_name}.Client documentation]"
            f"({service_name.doc_link}.Client)"
        ),
    )

    shape_method_map = shape_parser.get_client_method_map()
    for method_name, public_method in public_methods.items():
        if method_name in shape_method_map:
            method = shape_method_map[method_name]
        else:
            method = parse_method("Client", method_name, public_method)
        method.docstring = (
            f"[Client.{method_name} documentation]"
            f"({service_name.doc_link}.Client.{method_name})"
        )
        result.methods.append(method)

    service_model = client.meta.service_model
    client_exceptions = ClientExceptionsFactory().create_client_exceptions(
        service_model
    )
    for exception_class_name in dir(client_exceptions):
        if exception_class_name.startswith("_"):
            continue
        if not exception_class_name[0].isupper():
            continue
        result.exceptions_class.attributes.append(
            Attribute(
                exception_class_name, TypeClass(ClientError, alias="Boto3ClientError")
            )
        )

    result.attributes.append(
        Attribute(
            "exceptions",
            InternalImport(
                name=result.exceptions_class.name,
                module_name=ServiceModuleName.client,
                service_name=service_name,
            ),
        )
    )
    return result
